/** (c) Brendan Macmillan 2001, licensed under GNU's GPL 2 
		(If your project is not GPL, then a fee is due - see license on website)
    Website: http://www.csse.monash.edu.au/~bren/JSX
    Website: http://freshmeat.net/projects/jsx
    List (can read without subscribing, need to join to post):
      http://groups.yahoo.com/group/JSX-ideas/messages
		Commercial licensing enquiries only:
			bren@mail.csse.monash.edu.au */


/** ===========================================================================
	* ObjIn - (could be called XMLObjectInputStream)
	*	=====
	* Input/deserialization.
	* Class-specific readObject() and writeObject() serialization.
	* USAGE:
	* 	java JSX.ObjIn < test.xml
	*
	* There are many possible calls to the Serialization API; but they are
	* all callbacks, from within readObject(), and so in a sense are invisible.
	**/

//package ObjOutDev;
package JSX;
//import JSX.*;	//use this, not "package", if truly a "wrapper"
import java.io.*;	//Stream

public class ObjIn extends ObjectInputStream {
	//static final boolean DEBUG = false;
	static final boolean DEBUG = true;

	static final boolean TRACE = false;
	//static final boolean TRACE = true;

	static final boolean NONSTREAM = false;
	//static final boolean NONSTREAM = true;

	public static ClassLoader cl = null;


	/**	==========================================================================
		* CONSTRUCTORS
		* ============
		**/
	/** -------------------------------------------------------------------------
		* ObjIn(in) - core constructor (others call it).
		* ------
		**/
	final private ValidationList vlist = new ValidationList();
	private XMLDeserialize d;
	private Config config;	//we keep a reference to it, as well as XMLDeserialize
	public ObjIn(Reader in) throws IOException {
		super();	//sets flag, so readObject() defers to readObjectOverride()
		d = new XMLDeserialize(in);
		d.setArg(new Object[] {this});
			//d.readObjectArglist = new Object[] {this};
		if (TRACE) System.err.println("ObjIn(inReader)");
	}

	public ObjIn(Reader in, Config cfg) throws IOException {
		super();	//sets flag, so readObject() defers to readObjectOverride()
		config = cfg;	//for resolveClass
		if (cfg.importFile!=null) {
			try {
				XSLTimportI xsltImport = (XSLTimportI) Class.forName("JSX.XSLTimport").newInstance();
					//need an instance to invoke method (even tho it's a static method)
				in = xsltImport.preprocess(in, cfg);	//chain ahead
			} catch (ClassNotFoundException noXSLTimport) {	
				throw new IOException(noXSLTimport.toString());	//poor man's wrapping
			} catch (InstantiationException cannotConstruct) {
				throw new IOException(cannotConstruct.toString());
			} catch (IllegalAccessException cannotAccess) {
				throw new IOException(cannotAccess.toString());
			}
		}
		d = new XMLDeserialize(in, cfg);
		d.setArg(new Object[] {this});	//why not send this in via the constructor?
			//d.readObjectArglist = new Object[] {this};
		if (TRACE) System.err.println("ObjIn(inReader)");
	}

	/** -------------------------------------------------------------------------
		* ObjIn(in)
		* ------
		* Convenience: wraps as a Reader.
		**/
	public ObjIn(InputStream in) throws IOException {	//concession to convenience
		this((Reader)new InputStreamReader(in));
		if (TRACE) System.err.println("ObjIn(in)");
	}
	public ObjIn(InputStream in, Config cfg) throws IOException {
		this((Reader)new InputStreamReader(in), cfg);
		if (TRACE) System.err.println("ObjIn(in)");
	}

	/** -------------------------------------------------------------------------
		* ObjIn()
		* ------
		* Default to System.in (so can redirect files in etc).
		*	Incidentally: SecurityException is RunTime, so needn't explicitly throw.
		**/
	public ObjIn() throws IOException, SecurityException {
		this(System.in);		//default
		if (TRACE) System.err.println("ObjIn()");
	}

	public ObjIn(Config cfg) throws IOException, SecurityException {
		this(System.in, cfg);		//default
		if (TRACE) System.err.println("ObjIn()");
	}


	/**	==========================================================================
		* METHODS
		*	=======
		**/
	/** -------------------------------------------------------------------------
		* reset()
		* -------
		*	To reset alias table, mainly.
		* Not needed: resets automatically upon each new invocation.
		* Consider: option to change this when common objects are repeatedly
		* 	messaged.
		* Question: is this in OIS API?
		**/
	public void reset() {	d.reset(); }	//for aliases
	public void close() {
		d.close(); // should clear aliases, too
		// vlist.clear();	//already clear on our model, if doCallbacks called...
	}

	/** -------------------------------------------------------------------------
		* readObjectOverride() -- different arg/return form output version
		* --------------------
		*	The heart.  ObjectInputStream.readObject() defers to this, if flag set.
		* Consider: only IOException should be thrown.
		**/
//maybe should make a general loader for this stuff?  Eg MagicClass routines?

// Comment out, because we do in Java now
/*
	static Object vlist;
	static {
		try {
			Class validClass = Class.forName("java.io.ObjectInputStream$ValidationList");
				// vlist = magicClass.newInstance(validClass);	//create it
			java.lang.reflect.Constructor cons = validClass.getDeclaredConstructor(new Class[0]);
			cons.setAccessible(true);
			vlist = cons.newInstance(new Object[0]);
		} catch (Exception e) {
			//e.printStackTrace();	// SILENT, simply means "not 1.4"
		}
	}
*/

// Comment out old doValidation, because now we do it all in Java.
/*
	static java.lang.reflect.Method doValidationsMethod;
	static { //class initializer
		if (vlist!=null) {	//should be 1.4... should move to MagicClass
			try { //1.4 - also expect vlist to be non-null
				doValidationsMethod = vlist.getClass().getDeclaredMethod("doCallbacks", new Class[]{});
			} catch (NoSuchMethodException ee) {	// means "not 1.3 either"
				ee.printStackTrace();	//fail silently
				System.err.println("Couldn't invoke doCallbacks");
			}
		}

		if (doValidationsMethod==null) {	//so, it wasn't 1.4... let's try 1.3
			try { //1.3
				doValidationsMethod = java.io.ObjectInputStream.class.getDeclaredMethod("doValidations", new Class[]{});
			} catch (NoSuchMethodException e) {
				e.printStackTrace();	//fail silently
					// hyp: this was printed for Alex in j9
				System.err.println("--- J9?  Unable to set-up doValidationMethod ---");
			}
		}

		if (doValidationsMethod!=null) {	// if we couldn't find it (eg j9)
			doValidationsMethod.setAccessible(true);
				// before if, this caused a NullPointerException for Alex in j9
		} else {
			System.err.println("--- J9?  Unable to set-up doValidationMethod ---");
		}
	}
*/

	protected Object readObjectOverride() throws OptionalDataException,
                                               ClassNotFoundException,
                                               IOException {
		try {
			if (!d.superVersion) {	//only for non-super version
				if (d.invocationCount>0)	//ie if already inside
					defaultReadObject();	//false = don't read in fields	hmmmm....????
			}
			if (TRACE) System.err.println("readObjectOverride()");
			Object o = null;
			o = d.deserialize();	// if an error thrown from here, still need to clear
			if (TRACE) System.err.println("Object returned:" + o);
	
			if (d.invocationCount==0)	//only if *not* inside (ie if really finished)
				vlist.doCallbacks();	// clears list - JOS keeps open, in case more
/*
			try {	//only at *very* end - not in a nested readObject call?
				recursionDepthField.setInt(this, 0);	//just a hack, to use superclass
				if (vlist==null) //1.3
					doValidationsMethod.invoke(this, new Object[]{});	//and that's it!
				else
					doValidationsMethod.invoke(vlist, new Object[]{});	//and that's it!
				//} catch (java.io.NotActiveException notOn) {
				//	notOn.printStackTrace();
				//fail silently for 1.4 tests at the moment, for Shine Lu
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			} catch (java.lang.reflect.InvocationTargetException e) {
				e.printStackTrace();
			//see java pattern for this (from ObjectInputStream... should make it
			//into a utility method)
			} catch (NullPointerException e) {
				e.printStackTrace();	//fail silently
				System.err.println("attempt to doValidationsMethod.invoke()");
			}
*/
			return o;
		} finally {
			vlist.clear();	// odd: JOS keeps priorities open between streams...
		}
	}

	/** -------------------------------------------------------------------------
		* defaultReadObject()
		* -----------------
		* Not implemented: presently, default stuff is written anyway.
		**/
	public void defaultReadObject() throws IOException, ClassNotFoundException {
		if (TRACE) System.err.println("defaultReadObject()");
		//This does nothing!!
		//What if a class *depends* on these values *not* being read in?
		//Can we simulate the effect by not writing if defaultWriteObject() not
		//called?  That would help!!  Classes then *shouldn't* create this problem!

		d.defaultReadObject();	//forward it.
	}

	public void readFully(byte[] dst) throws
	//		ClassNotFoundException,
			IOException {
			//the "fully" is to do with blocking, not the array being full
	
		byte[] src = null;
		try {
			src = (byte[]) readObject();
		} catch (ClassNotFoundException e) {
			//it would be pretty serious if "byte[].class" could not be found!
			//This could happen if some other class...
			throw new IOException("Reading for a byte[]; but: "+e);	//rethrow corked
		}
		 //byte[] src = (byte[]) readObjectOverride();	//so needn't catch exception
		System.arraycopy(src, 0, dst, 0, src.length);	//err if dst too short
		//this is the odd thing you have to do in Java, to return by reference.
		//it would be nice if we could point the dst at the src instead...
		//probably *can* do this with the private native trick... ;-)
	}


	/** -------------------------------------------------------------------------
		* internalDefaultReadObject()
		* -----------------
		* Re-throws ClassNotFound as ClassNotFoundInDefault.  This subclasses
		* IOException, to fit the sig of readInt etc.  JSX requires this hack
		* because of the funny way it uses defaultReadObject to set up the
		* custom fields.
		**/
	void internalDefaultReadObject() throws IOException {	//for readInt etc
		if (d.superVersion) //not needed for superVersion
			return;	//ineff to factored out here, but in one place
		try {
			d.defaultReadObject();	//forward it.
		} catch (ClassNotFoundException e) {	
			throw new ClassNotFoundInDefaultException(e.getMessage());
		}
	}
	class ClassNotFoundInDefaultException extends IOException {
		ClassNotFoundInDefaultException(String msg) { super(msg); }
	}

	public int readInt() throws IOException {
		internalDefaultReadObject();	//false = don't read in fields
		return Integer.parseInt(d.getPrimString());
	}
	public float readFloat() throws IOException {
		internalDefaultReadObject();	//false = don't read in fields
		//return Float.parseFloat(d.getPrimString());
		return ParseUtilities.parseFloat(d.getPrimString());
	}
		//we are stricter than Java's own parse of Boolean...
		//(Suggest: This could go into ParseUtilities)
	public boolean readBoolean() throws IOException {
		internalDefaultReadObject();	//false = don't read in fields
		return ParseUtilities.parseBoolean(d.getPrimString());
	}
	public byte readByte() throws IOException {
		internalDefaultReadObject();	//false = don't read in fields
		return Byte.parseByte(d.getPrimString());
	}
	public char readChar() throws IOException {
		internalDefaultReadObject();	//false = don't read in fields
		//XML...
		return (char) Integer.parseInt(d.getPrimString());
		//return d.getPrimString().charAt(0);
		//return ParseUtilities.decodeXML(d.getPrimString());
	}
	public double readDouble() throws IOException {
		internalDefaultReadObject();	//false = don't read in fields
		//return Double.parseDouble(d.getPrimString());
		return ParseUtilities.parseDouble(d.getPrimString());
	}
	//Float - see above
	//Int - see above
	public short readShort() throws IOException {
		internalDefaultReadObject();	//false = don't read in fields
		return Short.parseShort(d.getPrimString());
	}
		//Also thanks to Andy Thomas - odd that both were left out!
	public long readLong() throws IOException {
		internalDefaultReadObject();	//false = don't read in fields
		return Long.parseLong(d.getPrimString());
	}

/* deprecated in java.io.ObjectInputStream
	public String readLine() throws IOException {
		System.err.println("readLine() not implemented");
		return null;
	}
*/
	public String readUTF() throws IOException {
		internalDefaultReadObject();	//false = don't read in fields
		//System.err.println("getPrimString() is correct, not readObject()");
		return d.getPrimString();
/*
		try {
			return (String) readObject();  //if it's not String, then an error!
		} catch (ClassNotFoundException e) {
			e.printStackTrace();	//shouldn't be needed: it should be a String.
		}
		return null;
*/
	}

	/**	--------------------------------------------------------------------------
		* Non-stream API
		* --------------
		* This is for stuff that is not actually reading (or writing, for ObjOut).
		* Eg: validation
		**/

/** registerValidation: we simply let this fall through to the superclass
	* implementation.  All we need do is call "doValidation()" when the entire
	* object graph is done.
	* (1). Where do we put it, so we know when it is done?
	* (2). How do we call it with reflection?
	*/
// Comment out, and do it in Java now
/*
	static java.lang.reflect.Field recursionDepthField;
	static { //class initializer
		try {	//should do in MagicClass; or dup code myself
			recursionDepthField = java.io.ObjectInputStream.class.getDeclaredField("depth");
		} catch (NoSuchFieldException e) {
			//e.printStackTrace();	//SILENT 1.4 fail: it expects a "depth" field
			try {
				recursionDepthField = java.io.ObjectInputStream.class.getDeclaredField("recursionDepth");
			} catch (NoSuchFieldException ee) {
				ee.printStackTrace();	//1.3 fail: it expects a "recursionDepth" field
			}
		}
		recursionDepthField.setAccessible(true);
	}

	{ //instance initializer - need an instance of field, to set it
		try {
			recursionDepthField.setInt(this, 1);	//just a hack, to use superclass
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (NullPointerException e) {
			//e.printStackTrace();	//fail silently
		}
	}
*/


	public void registerValidation(ObjectInputValidation obj, int prio) throws InvalidObjectException, NotActiveException {
		//throw new RuntimeException("registerValidation() invoked - not implemented");
		if (d.invocationCount==0)	//ie not active
			throw new NotActiveException("JSX.ObjIn: readObject not Active");
		vlist.register(obj, prio);

/*
		else {
			try {
				try { //1.4 - better to detect, and do within MagicClass - or dup self
					JSX.magic.MagicClassI magicClass = JSX.magic.MagicClassFactory.newInstance();


					java.lang.reflect.Field f = java.io.ObjectInputStream.class.getDeclaredField("vlist");
					magicClass.setObjectFieldValue(this, f, f.getType(), vlist);
				} catch (Exception problem) {
					//problem.printStackTrace(); //SILENT means not 1.4
				}
				super.registerValidation(obj, prio);	//forward it
			} catch (java.io.NotActiveException notOn) {
					//sh fail silently for 1.4 tests at the moment, for Shine Lu
				throw new InternalError("Stream not active");
			}
		}
*/
	}


	protected Class resolveClass(ObjectStreamClass v) throws IOException, ClassNotFoundException {
			//return Class.forName(v.getName(), true, this.getClass().getClassLoader());	//Wolf test: same as forName(name)
			//return Class.forName(v.getName(), false, cl);
		String classname = v.getName();
		if (config!=null && config.refactor!=null)
			classname = config.refactor.mapClassname(classname);
		Class clazz = null;
		try {
			//clazz = Class.forName(v.getName());	//this *would* solve Wolf's problem
			clazz=Thread.currentThread().getContextClassLoader().loadClass(classname);  //this brilliant bug-fix courtesy Ing Thomas Margreiter, 13 June 2002

		} catch (ClassNotFoundException e) {
			throw new ClassNotFoundException("forName(\""+v.getName()+"\") didn't find the class. Please report this problem. Wrapped: "+e.getMessage());
			//Will it work, in general?  We'll see...
			//clazz = super.resolveClass(v); //best: but latestUserDefinedLoader...
		}
		return clazz;
	}

	protected  Object resolveObject(Object obj) {
		throw new RuntimeException("resolveObject() invoked - not implemented");
	//	return null;
	}
	public int skipBytes(int len) {
		throw new RuntimeException("skipBytes() invoked - not implemented");
	//	return 0;
	}

/** communicate with XMLDeserilize */
	public ObjectInputStream.GetField readFields() throws
			IOException,
			ClassNotFoundException {
		defaultReadObject();	//false = don't read in fields
		if (TRACE) System.err.println("in.readFields(); invoked - about to return");
		return (GetField)d.getCurrentGetField();
			// we could even send in "d", to get the current class, instead of
			// making it non-static
	}
 


	/**	==========================================================================
		* TEST CODE
		* =========
		**/
	public static void main(String[] args) throws Exception {
		
		InputStream fin = System.in;
		if (args.length!=0) { //a filename
			fin = new FileInputStream(args[0]);
		}
/*
    Config cfg = new Config().setFormatted(true).setNoSuchFieldHandler(
      new NoSuchFieldHandler() {
        public String noSuchField(Class c, String fieldName) {
          System.err.println("--------wanting to map field "+fieldName);
          return fieldName.substring(1);
        }
      });
    cfg.setNoSuchFieldHandler(new NoSuchFieldHandler());	//default, so gives
*/

    Config c = new Config().setFormatted(true).aliasID(false);	// turn aliases off for eg
    //Config cfg = new Config().setFormatted(true).superVersion(true);

		//errors

		ObjIn in = new ObjIn(fin, c);
		//ObjIn in = new ObjIn(fin);
    //Config cfg2 = new Config().superVersion(true);
    //Config cfg2 = new Config().superVersion(false);
		//ObjOut out = new ObjOut(cfg2);
		ObjOut out = new ObjOut(c);
		Object o;
			//What if the object returned *is* null?
		while (true) {
			try {
				o = in.readObject();
				if (DEBUG) {	//sep debug info from deserialization and serialization
					System.err.println();
					System.err.println("Finished reading--------");
					System.err.println();
				}
				out.writeObject(o);
				out.flush();
				//NB: The Config object should work as an extension to the object,
				//that does not work through inheritance (and thus has a different
				//namespace).
											
				if (DEBUG)
					System.err.println("Finished writing--------");
			} catch (EOFException e) {	//simply end of file.
				break;
			} catch (Exception e) {
				System.err.println(e);
				e.printStackTrace();
				continue;
			}
		}
		out.close();
		in.close();
		System.exit(0);	//in case one of the deserialized objects started a thread
	}
}
