/** (c) Brendan Macmillan 2001, licensed under GNU's GPL 2 
		(If your project is not GPL, then a fee is due - see license on website)
    Website: http://www.csse.monash.edu.au/~bren/JSX
    Website: http://freshmeat.net/projects/jsx
    List (can read without subscribing, need to join to post):
      http://groups.yahoo.com/group/JSX-ideas/messages
		Commercial licensing enquiries only:
			bren@mail.csse.monash.edu.au */


/** ===========================================================================
  * XMLSerialize (JSX0.8.21).
	* ============
	* 
	* NB: This file best viewed with tabs set to 2.
	* For vim (vi clone):
:set ts=2
	* For emacs:
M-x edit-tab-stops
	* and then put a ":" in every second position (use space to overwrite).
	* (M is "Meta key"; usually just Alt)
	*
	* DESCRIPTION
	* Pass in an object, and it writes out a serialized version, in XML format.
	* It mimics Java's own serialization, and so can handle almost all classes,
	* including those with:
	*		- Private fields
	*		- A lack of a no-argument constructor
	*		- Their own internal serialization (except for Externalizable)
	*	It handles:
	*		- Circular references
	*		- Multi-dimensional arrays (ie arrays of arrays)
	*
	* For comments of any kind
	* (eg bug reports, bug fixes, suggestions, patches, criticism, comments,
	* insights, ideas, frustrations, irritations, or alternative software in the
	* same space):
				bren@csse.monash.edu.au  <G. Brendan Macmillan>
	**/

/** BUGS.
	*		Wrapper classes only serialized correctly when within container types
	*			- Bug fixed or not?
	*			- would it now work automatically?
	*		String as first object?
	**/

/** The mapping of Java to XML.
	*		Primitives map to attributes.
	*		References map to elements:
	*		- The element name is the *runtime class*.  eg <java.util.Hashtable>
	*		- A NAME_TOKEN (eg "obj-name") attribute is the name of the field, if it
	*		has a name.
	*		- NB: NAME_TOKEN is illegal as a Java field (because of '-'), so no
	*		namespace conflict.
	**/

/** Logic flow.
	*		For each class: (serializeObject)
	*			write all fields as attributes - if primitive (serializeFields)
	*			write all fields as elements - if reference (serializeReferences)
	*				-> recurse
	**/

/**	Notes.
	* NB: Newlines within strings are outputted *exactly* as they are stored
	*	in the string.  This might not be what you want, when moving across DOS
	* (/n/r) and UNIX (/n).
	**/

/** To do.
	*		- nice, informative error messages when the field is not found.
	**/

/** Circular References.
	*	Increment a count whenever an object is serialized - the aliasSerialNumber.
	*
	* This number becomes the alias (or alternate name) for that object.
	*	It is the *very* simplest implementation; not very human readable, but
	* will reveal implementation difficulties without complicating them.
	* NB:
	*		- A null value is not an object (so doesn't increment the count)
	*		- "String ambivalence": String are objects, but encoded as attributes
	*		Could just say <fieldname>-alias="<aliasname>"
	*		Similarly, could say <fieldname>-null="" for null (!= to empty string!)
	*		Difficulty is extra parsing time to detect the "-".  OK if can do it
	*			efficiently, which would be good, anyway.
	*		- A normal object is a an object.
	*		- A wrapper object is an object.
	*
	* In present implementation, aliases are needed in
	*		- serializeObject,
	*		- serializeWrapper,
	*		- serializeFields - for strings (strings in containers are in wrappers)
	*	NB: it is pre-inc.  Begins at -1, and is PRE-INC just before each use.
	**/

/** Handling internal serialization (ObjOut.java ObjIn.java).
	*	(1). When a class specifies its own serialization, it only does it for
	*	itself, not its ancestors (superclasses).  Therefore, the test for a
	*	writeObject() method must be in getAllFields(), which walks up the
	*	inheritance tree.
	*	(2). This is presently called twice for each object: once for primitives,
	*	and once for objects (references).  This is because JSX encodes these
	* very differently in XML (attributes vs. tags), though this is not the
	*	only implementation.  We now pass in a flag so that the internal
	*	serialization logic is only executed once (in the first call, which comes
	* when serializing primitives (as attributes).
	* (3). A class's internal serialization invokes the whole process again
	* for objects (references), using writeObject() (in ObjOut.java),
	* with no parameters to indicate that it is not a fresh level serialization.
	* This is important to know, for reseting the alias table, for correct
	* tabbing in formatting, and for whether to save up the Objects until
	* after the primitives are done, instead of doing them right away.  These
	* two cases of fresh vs. internal are distinguised by:
	*		(a). serialize invocation (not serializeObject, which nested tags invoke)
	*		(b). invocationCount != 0
	*	The primitive writes, such as writeInt() are easily distinguished, as that
	* is the only way they may legitimately be invoked (for JSX, anyway).
	*	(4). Thus, the internally serialized objects and primitives are gathered
	*	and collected in a bucket.
	*	(5). For primitives, it would be nice for this bucket to subclass Field,
	* so they could remain invisible to the primitive-handling method
	*	serializeFields().  However, Fields are "final", so instead we pass back
	*	an array of type Object, and serializeFields() must distinguish between
	*	them with the instanceof operator, a name, value String pair (String[]?)
	*	Or class x (String name; String value;}?  putField() is no good.
	*	(6). NB: No recursion occurs from this internal serialization - primitives
	* don't recurse, and the object references are merely noted, not written
	* at this point.  Thus, the invoked writeObject() is guaranteed to return
	*	immediately.
	* (7). Thus, we can use the stored primitives and objects immediately (like
	*	a one-level stack).   The primitives are written as above; and the 
	*	object references are serialized just before the call to
	* serializeReferences.
	* (8). NB: This call to serializeReferences() will invoke getAllFields() a
	*	second time, and we don't wish to invoke the internal serialization again,
	* or it will be awkward to distinguish the first and second lot of stored
	*	objects.  However, [NB] when we implement defaultWriteObject(), we will
	*	still need to detect (but not invoke), so we don't serialize twice.
	* (9). We will need to start fresh for each invocation of an internal
	* serialization, so mix up the storage of primitives and object references
	* from different classes.  Note that this can be true even within the one
	* object, [NB] when we implement internal serialization of ancestor
	*	superclasses (which will be needed we when use Hashtable's internal
	* serialization instead of the present ad hoc approach, and we want to allow
	* users to subclass it - as per Michael Oswall's request.
	**/

package JSX;
import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import java.util.Hashtable;	//for Hashtable
import java.util.LinkedHashMap;	//for Hashtable
import java.util.Map;

public class XMLSerialize {
	static final boolean defWO_DIRECT = false;
	//static final boolean defWO_DIRECT = true;

	static final boolean DEBUG = false;	
	//static final boolean DEBUG = true;

	static final boolean DEBUG_HASH = false;	//Hashtable uses .equals(), not ==
	//static final boolean DEBUG_HASH = true;

	static final boolean TESTCIRC = false;
	//static final boolean TESTCIRC = true;	//circular references

	static final boolean INTERNAL_DEBUG = false;
	//static final boolean INTERNAL_DEBUG = true;	//internal serialization

	static final boolean CACHE_DEBUG = false;
	//static final boolean CACHE_DEBUG = true;//messy multi invoke of getAllFields

	static final boolean REENTRY_DEBUG = false;
	//static final boolean REENTRY_DEBUG = true; //when serializeObject() called

	static final boolean PUT_FIELD_DEBUG = false;
	//static final boolean PUT_FIELD_DEBUG = true;

	static final boolean OPT_DEBUG = false;
	//static final boolean OPT_DEBUG = true;


	static final int ALIAS_INIT = -1;//so pre-increment in serializeObject
	private int aliasSerialNumber;	//simplest circular reference names

	static final String ALIAS_TAG_TOKEN = "alias-ref";
	static final String ALIAS_ATTR_TOKEN = "alias";	//'-' not needed, b/c alone
	static final String ALIAS_STRING_TOKEN = "-alias";

	static final String VERSION = "Version: JSX0.8.13 (maybe out of date - sorry! See the name of the jar.)";	//MUST remember to update this!

	/**	
		*	"-" avoids namespace conflicts between serialization flags and 
		*	actual class or field names.
		*	"-" legal XML attr; illegal java field; illegal java class
		**/
	//static final String NAME = "is-the";	//"-" legal attr; illegal field
	//static final String NAME = "class-of";	//opposite of "instanceof"
	//static final String NAME = "run-type-of";	//opposite of "instanceof"
		//there is an issue with compatibility here!  I keep changing it, and old
		//serializations are no longer compatible with it.
	static final String NAME_TOKEN = "obj-name";

	static final String CLASSNAME_TOKEN = "className";
	//static final String ARRAY_TOKEN = "A-";
	//static final String ARRAY_TOKEN = "[]";	//not allowed as XML names!
	static final String ARRAY_TOKEN = "ArrayOf-";
	//static final String ARRAY_PRIMITIVE_INDEX_TOKEN = "a-";
		//NB: prefixes needed - attr names can't start with a numeral
	static final String ARRAY_PRIMITIVE_INDEX_TOKEN = "a";	//'-' not needed
	static final String INTERNAL_PRIMITIVE_TOKEN = "_.";	//'-' not needed
		//NB: '_' is perfect to convey the sense of an internal name!
		//This may also be appropriate for arrays; but 'a' for array, and no
		//	other attributes to confuse (except name and length)
		//Problem: a valid java fieldname.  Although _0 _1 _2 (ie with numerals)
		//	are unlikely to be used as fieldnames, the simplist parsing is
		//	confused by fieldnames such as "_name".  Using the prefix "_."
		//	should fix this.  This bug was discovered by Thomas Hiller, who
		//	thus became the third successful JSX Challenger, on:
		//	Tue, 16 Jan 2001 17:58:24 +0100

	static final String LENGTH_TOKEN = "length";
	static final String VALUE_TOKEN = "valueOf";
	static final String BINARY_DATA_TOKEN = "binary-data";
	static final String NULL_TOKEN = "null";

	//static final String SUPER_TOKEN = "super";	//a reserved keyword
	//static final String SUPER_TOKEN = "subclass";	//**reversed** order
	static final String SUPER_TOKEN = "sub-class";	//**reversed** order
	//static final String SUPER_TOKEN = "extended-by";	//reversed order
	static final String OPT_DATA_TOKEN = "opt-data";
	static final String OPT_PRIM_DATA_TOKEN = "opt-prim-data";

	static final String ALIAS_ID_TOKEN = "alias-ID";


	static final String SPACER = "  ";		//\t

	static final String JSX_HEADER_TARGET = "jsx";	//mv to constants interface
	static final String XML_HEADER_TARGET = "xml";

	static final int NOT_SERIALIZED =
		Modifier.STATIC | Modifier.TRANSIENT;
//		| Modifier.FINAL; //todo 26June2001: be able to set final fields


	public static void main(String ars[]) {
		System.err.println(VERSION);
	}
	/**--------------------------------------------------------------------------
		* API
		* ===
		* CONSTRUCTORS
		* ============
		**/
	//ALIAS ID related:
	private boolean stringAsElement = true;	//for alias-id
	//private boolean stringAsElement = false;	//for alias-id

	//should be ganged to the above
	private boolean includeAliasID = true;	//for alias-id
	//private boolean includeAliasID = false;	//for alias-id


	private boolean superVersion = false;	//regression test
			//config can override!!!
	//private boolean superVersion = true;

	private boolean optVersion = superVersion;	//gang for convenience
	//private boolean optVersion = false;
	//private boolean optVersion = true;
	private boolean primOrderVersion = optVersion;	//gang for convenience
	//private boolean primOrderVersion = false;
	//private boolean primOrderVersion = true;

	private PrintWriter out;
	public XMLSerialize(PrintWriter out) throws IOException {
		this(out, new Config());	// so Config becomes the default
		// this.out = out;
	}
	Config cfg = new Config();	//so never null, and Config controls defaults
			//we cut and pasted constructors - better to call the other
				//if none call, same as cfg=null (is it?  Does it matter?)
	public XMLSerialize(PrintWriter out, Config cfg) throws IOException {
		if (cfg==null) throw new IOException("Config object must not be null");
		this.cfg = cfg;
		setFormatted(cfg.formatted);
		superVersion = cfg.superVersion;
		optVersion = superVersion;	//gang for convenience
		primOrderVersion = optVersion;	//gang for convenience
		stringAsElement = cfg.aliasID;	//alias ID does several things
		includeAliasID = cfg.aliasID;	//alias ID does several things
		this.out = out;
	}
	public XMLSerialize(Config cfg) throws IOException {	//defaults to System.out
		this(new PrintWriter(new OutputStreamWriter(System.out), true), cfg);
	}
	public XMLSerialize() throws IOException {	//defaults to System.out
		this(new Config());	// so Config becomes the default
			// *and* we run through the code that sets it
		// this(new PrintWriter(new OutputStreamWriter(System.out), true));
	}


	//legacy method
	public void serialize(Object o, PrintWriter newOut) throws IOException {
		out = newOut;	//can change output destination between invocations
		commonSer(o);
	}
	public void serialize(Object o) throws IOException{
		commonSer(o);
	}
	/** ------------------------------------------------------------------------
		* commonSer()
		* -----
		* Heart.
		* Consider: should this be static?
		**/
		 /* Mechanism for internal serialization to pass out objects.
			* Consider: reimplementing with faster ArrayList (not synch).
			* Danger: incremental capacity increases.
			* Consider: doubling concept used in OOS.
			*/
	int invocationCount = 0; //don't ++ in nest, so max value is 1...
	private void commonSer(Object o) throws IOException{
		//nested invocation - custom serialization of a class's writeObject()
		if (invocationCount!=0) {		//Exception if no objStore exists.
			if (superVersion)
				serializeObject(o, null, "");		//indent unknown - need a stack!
			else
				objStore.add(o);		//for internal serialization.  NB: doesn't recurse.
			return;
		//first invocation (not nested)  [ED: should this be first?]
		} else {
			reset();
			invocationCount++;
			String jsxVersion;
			if (superVersion)
				jsxVersion = "2";
			else
				jsxVersion = "1";
			toOutPrintln("<?"+JSX_HEADER_TARGET+" version=\""+jsxVersion+"\"?>");

			serializeObject(o, null, "");
			toOutPrintln();
			if (TESTCIRC) {
				System.err.println("Count was: " + (aliasSerialNumber-ALIAS_INIT));
				System.err.println();
			}
			invocationCount--;
			if (cfg.dtdGenerator!=null) cfg.dtdGenerator.generate(); //hit it, man.
		}
	}



	/** ------------------------------------------------------------------------
		* METHODS
		* =======
		**/
	/** ------------------------------------------------------------------------
		* reset()
		* -----
		**/


	static class MapProtect extends java.util.Stack {
		void protect(Map[] cache) {
			push(cache);
		}
		Map[] retrieve() {
			return (Map[])pop();
		}
		Map[] value() {
			return (Map[])peek();
		}
	}

	static class MapCatcher extends java.util.Stack {
		//static final boolean DEBUG = true;
		static final boolean DEBUG = false;

		void newCatcher() {	// new frame or context
			push(null);	// begin with a default value
			if (DEBUG) System.err.println("newCatcher: "+this);
		}
		void set(Map map) { // replace top with new value
			pop();	// note: if Boolean was mutable, we would just change it
			push(map);
			if (DEBUG) System.err.println("set(): "+this);
		}
		Map freeCatcher() {
			if (DEBUG) System.err.println("freeCatcher: "+this);
			return (Map)pop();	// discard catcher
		}
		public String toString() {
			if (size()==0) return "0";
			else return size()+", "+peek().toString();
		}
	}



/** (Conventional) use of a stack to save state (normalprotection: "pull"
 * communication: the push does not put information on the stack, but a holder
 * into which information can be caught.  Once sent, the information is
 * retrieved with "pop" (or "is").  The metaphor is a hand extended (push) to
 * "catch" the information, and then to bring it back (pop).
	**/
	static class BooleanProtect extends java.util.Stack {
		void protect(boolean[] cache) {
			push(cache);
		}
		boolean[] retrieve() {
			return (boolean[])pop();
		}
		boolean[] value() {
			return (boolean[])peek();
		}
	}


/** (Unconventional) use of a stack for communication: the "push" extends a a
 * holder in which information can be caught.  The "pull" brings it back.  The
 * metaphor is a hand extended to "catch" and then bring back the information.
	**/
	static class BooleanCatcher extends java.util.Stack {
		//static final boolean DEBUG = true;
		static final boolean DEBUG = false;

		void newCatcher() {	// new frame or context
			push(Boolean.FALSE);	// begin with a default value
			if (DEBUG) System.err.println("newCatcher: "+this);
		}
		void set(boolean input) { // replace top with new value
			Boolean b = new Boolean(input);
			pop();	// note: if Boolean was mutable, we would just change it
			push(b);
			if (DEBUG) System.err.println("set(): "+this);
		}
		boolean is() { // read top (not a pop)
			if (DEBUG) System.err.println("is(): "+this);
			return ((Boolean)peek()).booleanValue();		// if True, then true; else false
		}
		boolean freeCatcher() {
			if (DEBUG) System.err.println("freeCatcher: "+this);
			return ((Boolean)pop()).booleanValue();	// discard catcher
		}
		public String toString() {
			if (size()==0) return "0";
			else return size()+", "+peek().toString();
		}
	}

	public void reset() {	//can reset - important part of API!
		if (invocationCount!=0)
			throw new Error("Serialize: Attempt to reset alias table mid-way.");
		defaultCatcher.clear();
		putFieldObjectsCatcher.clear();
		primStore = new Vector();
		initAlias();
		putFieldClassStack.clear();
	}
	/** ------------------------------------------------------------------------
		* defaultWriteObject()
		* ------------------
		* Forwarded from ObjOut.
		* NB: no object specified.  It is merely returning a flag.
		**/

	private BooleanCatcher defaultCatcher = new BooleanCatcher();
	private BooleanProtect defaultProtect = new BooleanProtect();

	private MapCatcher putFieldObjectsCatcher = new MapCatcher();
	private MapProtect putFieldObjectsProtect = new MapProtect();
	private Vector primStore;	// = new Vector();
	Vector objStore;	//NB: a new one created for each
	public void defaultWriteObject() throws IOException {
		if (DEBUG)
			System.err.println("defaultWriteObject() in XMLSerialize: should store it");
		defaultCatcher.set(true);
		if (defWO_DIRECT) {
			System.err.println("Object "+thisObjectStack.get(thisObjectStack.size()-1));
			System.err.println("\tof class "+clazzStack.get(clazzStack.size()-1));
			System.err.println("\twith indent '"+indentStack.get(indentStack.size()-1)+"'");
			System.err.println("\tJust called defaultWriteObject");
			System.err.println();
			serializeReferences(
				thisObjectStack.get(thisObjectStack.size()-1),
				(Class)clazzStack.get(clazzStack.size()-1),
				(String)indentStack.get(indentStack.size()-1) );
		}

	}	
	public void flush() { out.flush(); }
	public void close() { out.flush(); out.close(); }

	/** ------------------------------------------------------------------------
		* writePrim(str)
		* ------------------
		* Forwarded from ObjOut.
		* Caller (ObjOut) must convert each of the primitive types to String
		* first.  This is trivial eg: (x+"")
		**/
	public void writePrim(String str) {
		if (DEBUG)
			System.err.println("writeInt(i) in XMLSerialize: should store it");
		if (primOrderVersion) {
			//direct write-through - ordering is then perfectly correct
			printOpenTag(OPT_PRIM_DATA_TOKEN, "", false);	//indentation - not object
			printAttr(VALUE_TOKEN, ParseUtilities.encodeXML(str));
				//could be string - better to take care of that case earlier...
			toOutPrintln("/>");	//only prim fields
		} else{
			primStore.add(str);
		}
	}


	/**===========================================================================
		* alias nascent-object
		* --------------------
		* collect init into one routine... suggests it should be a class.
		* change to public, so ObjOut can access it
	* ASIDE: could do the alias "nascent object" as an inner class - and
	* implement it by subclassing (instead of wrapping), and use HashMap
	* (instead of Hashtable)
		**/
	private Hashtable aliasHash = new Hashtable();	//reuse the one
	private void initAlias() {
		aliasSerialNumber = ALIAS_INIT;
		aliasHash.clear();
	}
	/**--------------------------------------------------------------------------
		* put Alias
		* --------
		*	inline = !serialized;
		* if (inline)
		*		//add to hash
		*	else
		*		//write as alias
		*	return (inline);		//to tell caller to serialize it inline!
		*		//else, caller leaves the ref writing up to this routine
		**/
	//simplest way to circumnavigate the ".equals" instead of identity test 
	//of Hashtable (in Java 1.2)
	class IdObject {
		Object o;
		public int hashCode() {
			return System.identityHashCode(o);
		}
		public boolean equals(Object x) {	//essense
			return ((IdObject)x).o == this.o;	//assumption that used by Hashtable,
				//and that the tested part (key) will always be an IdObject
		}
		public IdObject(Object o) {
			this.o = o;
		}
	}
	boolean areAliasesUsed=true; //default "on"; set in constructor
		//this name feels confusing to me...
	//Return value: false = I didn't add it (already there); don't inline it.
	//true = created new alias' you need to serialize it inline.
	private boolean putAlias(Object o, String fieldName, boolean isStringAttr, String indent) {
		//if (!areAliasesUsed)
		//	return true;
		if (DEBUG_HASH) System.err.println("  entering putAlias with "+o);

		if (o==null) return true;	//carry on
		Object alias;	//will be String, but save the conversion
			if (DEBUG_HASH) {
				for (Enumeration keys=aliasHash.keys(); keys.hasMoreElements();) {
					Object a = keys.nextElement();
					System.err.println(a + " -- " + a.getClass() + " -- " +a.hashCode());
				}
				System.err.println("looking for: "+o+" -- "+o.getClass() + " -- " +System.identityHashCode(o));
				System.err.println("Got: "+myget(o));
				System.err.println("Same object? 1 " + (myget(o)!=null) );
				System.err.println("Same object? 2 " + (mycontainsKey(o)) );
			}
		if ( (alias = myget(o))==null) {
			++aliasSerialNumber;	//pre-inc (use this syntax to highlight it)
			myput(o, aliasSerialNumber+"");	//add (convert to String)
				if (DEBUG_HASH) System.err.println("  recording new alias: "+indent +"inline " +aliasSerialNumber+ ":");
			//Need to:
			//	indicate "inline" somehow.
			//	write as inline (same as now).
			return true;	//yes, should serialize the object inline
		}
		else
		{
			if (isStringAttr) {
				printAttr(fieldName + ALIAS_STRING_TOKEN, alias+"");
			}
			else {
					if (DEBUG_HASH) System.err.println("  repeating old alias: "+indent+ "alias " +alias+ ":");
				printOpenTag(ALIAS_TAG_TOKEN, indent, false);	//not a *fresh* object
				if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);	//nice to reuse
				printAttr(ALIAS_ATTR_TOKEN, alias+"");
				toOutPrintln("/>");		//empty tag - under control of caller
			}
			return false;	//no, don't serialize inline; we (or you?) write the alias
		}
	}
	private void removeAlias(Object key) {
		if (areAliasesUsed) {	//if aliases are used, we don't mess with them
		} else {	//else, we remove them, so only the path remains.
			try {
				Object oo = aliasHash.remove(new IdObject(key));
				if (oo==null)
					throw new Exception("Try to remove an object not in aliasHash: " +key);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}
	public Object myput(Object key, Object value) {
		if (cfg.dtdGenerator!=null) cfg.dtdGenerator.populate(key);	//the object
		return aliasHash.put(new IdObject(key), value);	//just a dumb wrapper
	}
	public Object myget(Object key) {
		return aliasHash.get(new IdObject(key));	//just a dumb wrapper
	}
	public boolean mycontainsKey(Object key) {
		return aliasHash.containsKey(new IdObject(key));	//just a dumb wrapper
	}



	/**--------------------------------------------------------------------------
		* Object (map to an Element)
		* ------
		* tag (name of element) is always the runtime class (type) of the object.
		* SUGGEST: write open and close tags separately; but problem of attributes
		*  check: Wrapper classes here (String, primitive wrappers)
		*  check: container classes (like Hashtable)
		*
		* NB: aliasSerialNumber++ three times, before of nature of sub-arrays.
		**/

	Class[] OOS_ARGS = {ObjectOutputStream.class};	//works fine!
		//just mispelling?  it was: "ObjectOutpuStream.class"
	Object[] writeObjectArglist;  //set by caller, using setArg below
	ObjectOutputStream thisOut;
	/**--------------------------------------------------------------------------
		* setArg()
		* -------
		* Consider: putting this in constructor; or caller in same package or same
		* class.
		* THIS IS TRICKY!!! MUST BE CHANGED!!!
		**/
	public void setArg(Object[] argList) {
		writeObjectArglist = argList;
		thisOut = (ObjectOutputStream)argList[0];
	}


/** ================================================== =========================
	* PutFieldImpl - inner class
	* ------------
	* Need not buffer, because we don't worry about the order (but JOS does).
	*
	* Need to package-qualify PutField, because not directly subclassing.

	* (1). Treat them as full objects, invoking their writeObject() etc
	* (2). Buffering fields?  Or is this class itself the buffer?
	*
	* DEFER:
	* Strings... aliasing them?
	* 	For now, do as objects.
	* 	-> Machine readable, but less humam readable.

	* Indentation
	* 	(the positional stuff doesn't actually recurse, but stores it (memory
	* 	overhead is only the path to the current node).
	* 	Recursing requires some communication, somewhere
	* 	-> Machine readable, but much less humam readable.

	* Shadowed names
	* 	(1). superclasses fields shadowed by us,
	*		(2). and those in subclasses that shadow us ourselves.
	* 	-> Not machine readable, or human readable; but rare.

	* Empty tag
	*		<tag/> for empty tag (currently always does <tag> </tag>
	*		-> Machine readable, and humam readable.  Mere finesse.

	* Nesting
	* 	my theory is that stacking it (same as for positional) is the
	* 	easiest short-term solution, because it fits in with the rest of the
	*		code most easily.
	* 	-> necessary for correctness
	**/
	boolean putFieldDiditAll;	//should stack - or call through an object...

	//static final class PutFieldImpl extends java.io.ObjectOutputStream.PutField 
	final class PutFieldImpl extends java.io.ObjectOutputStream.PutField {
		Map objMap = null;
		Map primMap = null; // store primitives as Strings
		Class holdingClass;
		Map serialMap;

		public PutFieldImpl() { //"public" - but not needed, despite javadocs
			if (PUT_FIELD_DEBUG) System.err.println("PutFieldImpl.<init>");
			holdingClass = (Class) putFieldClassStack.peek();
			serialMap = SerialPF.getFieldNameType(holdingClass);
/* NOTE: superVersion does *not* work with this!
	* The opt-tag is confused:
	            <opt-data>
 lowestSetBit="-2" firstNonzeroByteNum="-2" signum="1" bitLength="-1" bitCount="
-1">
*/
			if (PUT_FIELD_DEBUG) {
				System.err.println("Present class is: "+holdingClass);
				System.err.println("Valid fields are: "+serialMap);
				try {
					SerialPF.main(new String[]{holdingClass.getName()});	//show it
				} catch (Exception e ) {
					e.printStackTrace();
				}
			}

			// there seems to be no real reason to use a hash map, as we just
			// iterate, not search.  However, it does no harm, and we do want pairs.
			// NB: linkedlist map maintains the order!
			try {	// LinkedHashMap only available in 1.4
				objMap = (Map)Class.forName("java.util.LinkedHashMap").newInstance();
				primMap = (Map)Class.forName("java.util.LinkedHashMap").newInstance();
			} catch (ClassNotFoundException e) {
			} catch (InstantiationException e) {
			} catch (IllegalAccessException e) {
			}
			if (objMap==null || primMap==null) {	// neater than repeating in catch
				objMap = new HashMap();
				primMap = new HashMap();
			}

			// populate objMap & primMap with default values
			// NB: in JOS,  if no put(), still serialized with default value
			for(Iterator i = serialMap.entrySet().iterator(); i.hasNext();) {
				Map.Entry pair = (Map.Entry) i.next();
				Class type = (Class) pair.getValue();
				String fieldName = (String) pair.getKey();
				if (type.isPrimitive()) {
					if (type==boolean.class)
						primMap.put(fieldName, "false");
					else
						primMap.put(fieldName, "0");
				} else {	// must be a reference type (object or array)
					objMap.put(fieldName, null);
				}
			}
		}


	/** Write fields.
			 	<tag primPutField="primValue" _.primCustomData= "primValue">
					<objCustomData/>
					<objPutField obj-name="fieldName"/>
				</tag>
		**/
	/**
		* Primitives -> attributes.
		*  - *no* actual fields
		*  - custom data
		*  - PutField fields
		**/
	/**
		*	Objects -> elements.  [I keep wanting to say "objectives"]
		*  - *no* actual fields
		*  - custom data
		*  - PutField fields
		**/
		public void write(ObjectOutput objectOutput) throws IOException {
		//(1). Access to current state
		//----------------------------
			XMLSerialize s = ((JSX.ObjOut) objectOutput).s;
			String indent = " "; //IDEA: ++ on openTag entry; -- on closeTag exit?

		//(2). Attributes <- primitives
		//-----------------------------
			//(a). *NO* actual fields

			//(b). virtual fields (ie PutField)
			for(Iterator i = primMap.entrySet().iterator(); i.hasNext();) {
				Map.Entry pair = (Map.Entry) i.next();
				s.printAttr(pair.getKey()+"", ParseUtilities.encodeXML(pair.getValue()+""));
					//Downside is inefficiency. Ideally, only encodeXML if a char - but
					//how to know what the type is?
			}

			//(c). custom fields (primitives)
				//(cut and paste from serializeFields (out-factor it from here & there?)
			int i=0;
			for (Enumeration keys=s.primStore.elements(); keys.hasMoreElements();i++){
				s.toOutPrintln();							//Attribute indenting
				s.toOutPrint(indent+s.spacer);		//indent each attribute
				s.printAttr(INTERNAL_PRIMITIVE_TOKEN+i, ParseUtilities.encodeXML((String)keys.nextElement()));
				//NOTE: will never be String, because it is serialized using object.
			}
			s.primStore.clear();	//NB: reset when consumed
				//so isn't written out again by serializeFields

				// We comment out below, because we now leave the tag open, so other
				// primitive values can be appended.
			//s.toOutPrintln(">");	//we don't know the tag name. Easier to leave to
				//serializeObject, and *always* do </endTag>... we could communicate
				//more information through the same mechanism, of course (since we
				//already need to communicate if PutField handled it or not.
				//		ie: in putFieldDiditAll

		//(3). Elements <- objects
			//(a). *NO* actual fields

			//(b). virtual Object fields (ie PutField)
				// --> move next to serializeReferences
			putFieldObjectsCatcher.set(objMap);
/*
			for (Iterator iObj = objMap.entrySet().iterator(); iObj.hasNext();) {
				Map.Entry pair = (Map.Entry) iObj.next();
				s.serializeObject(pair.getValue(), pair.getKey()+"", indent);
					//NB: [name-value] swapped
			}
*/

			//(c). custom fields (primitives) <----- not yet implemented
				//(cut and paste from serializeObject (out-factor it from here & there?)
/*
			Vector localObjStore = objStore;//local->stack (recursion)
			Vector localDefaultCache = defaultCache;//local->stack (recursion)
			for (Enumeration keys=localObjStore.elements(); keys.hasMoreElements();) {
				Object a = keys.nextElement();
				if (INTERNAL_DEBUG && a!=null) System.err.println("in objStore: "+a+"\n"+a.getClass());
				serializeObject(a, a.getClass(), null, indent+spacer+spacer);	//debug print out
			}
				//objStore.clear(); //NB: reset when consumed
				//try: No need - now a new one each time
			if (INTERNAL_DEBUG) {
				System.err.println("should be cleared: "+localObjStore.size());
					//do internal first, so parser knows end automatically.
				System.err.println("references:");
						}
			defaultCache = localDefaultCache;	//restore! (count is OK)
*/

			if (PUT_FIELD_DEBUG) {
				System.err.println("PutFieldImpl.write(out)");
				System.err.println("objMap:");
				System.err.println(objMap);
				System.err.println("primMap:");
				System.err.println(primMap);
			}
//			s.putFieldDiditAll = true;	//this should really be a "push"!!
		//hyp: have a theory that this is not necessary...
		}

		private void check(String name, Class clazz) {
			//System.err.println(">>>>>>>>>>  CHECKING "+name+"; sfp/SF are: "+serialMap);
			if (PUT_FIELD_DEBUG) {
				System.err.println("\texpected "+clazz+", got " +serialMap.get(name));
			}
			if (!serialMap.containsKey(name))
				throw new IllegalArgumentException("no such field - " +name +" is not a valid 'Serializable Field' of "+holdingClass);
			// Can the field be assigned to from the value?
			// ie. is field a superclass of the value?
			// For primitives, is it the same?
			if ( !((Class)serialMap.get(name)).isAssignableFrom(clazz) )
				throw new IllegalArgumentException("no such field - "+name +" expects a "+serialMap.get(name)+" and cannot be assigned to from '"+ clazz+ "'");
		}

	/** put() for all 8 primitives, and 1 object */
		public void put(String name, boolean value) {
			check(name, boolean.class);
			primMap.put(name, value+"");
		}
		public void put(String name, char value) {
			check(name, char.class);
			primMap.put(name, value+"");
		}
		public void put(String name, byte value) {
			check(name, byte.class);
			primMap.put(name, value+"");
		}
		public void put(String name, short value) {
			check(name, short.class);
			primMap.put(name, value+"");
		}
		public void put(String name, int value) {
			check(name, int.class);
			primMap.put(name, value+"");
		}
		public void put(String name, long value) {
			check(name, long.class);
			primMap.put(name, value+"");
		}
		public void put(String name, float value) {
			check(name, float.class);
			primMap.put(name, value+"");
		}
		public void put(String name, double value) {
			check(name, double.class);
			primMap.put(name, value+"");
		}

	/** put() for objects */
		public void put(String name, Object value) {
			check(name, value.getClass());
			//BUG: what to do with null values?  Need to be able to record them...
			//JOS just uses an array (I wonder if this could be a problem with my
			//other stacks?  They are all Vectors, I think, so OK)

			/* JOS spec says object must be same type as field - not subtype
				we duplicate that functionality here, under protest */
			objMap.put(name, value);	//what could be easier?
		}

/*
		// route above calls through here
		private void _primPut(String name, String value) {
			// (1). how know which "clazz" we are in?
				// Map serialMap = SerialPF.getFieldNameType(clazz);
			// (2). if contains.
			primMap.put(name, value);
		}
*/

	}


//	private void writeObject(ObjectOutputStream oos) {
//	}
/**
	* SerializeObject(obj, frameClass, fieldName, indent)
	* ---------------
	* Recurses both on object and class.  Object is for aggregate objects
	* (contained by reference); Class is a frame to inspect a different part
	* of the superclass stack.  Typical (object) recursion:
			serializeObject(o, o.getClass(), fieldName, indent+spacer+spacer);
 */
	private void serializeObject(Object o, String fieldName, String indent)
			throws IOException {
		if (o==null) { //because of the o.getClass() below
			serializeObject(true, o, null, -1, fieldName, indent); //start recursion
			//serializeObject(o, null, fieldName, indent); //start recursion
		} else {
			Class[] superClasses = null;
			//serializeObject(o, o.getClass(), fieldName, indent);
			if (superVersion) {
				if (o.getClass()==Object.class ||		//special cases: ignore subclasses
						o.getClass()==Vector.class ||
						o.getClass()==Hashtable.class ||
						o.getClass()==Class.class ||
						o.getClass()==String.class
						)	{ //special case
					superClasses = new Class[] { o.getClass() };
				} else {
					superClasses = getReversedSuperClasses(o.getClass());//leaf; no Object
					//	o.getClass() is the reason for the o==null
				}
			}
			else {
				superClasses = new Class[] { o.getClass() };
			}
			serializeObject(true, o, superClasses, 0, fieldName, indent);
				//can't loop, need to recurse, because nested.
		}
	}

	private void serializeObject(boolean superHeader, Object o, Class[] superClasses, int superI, String fieldName, String indent) 	throws IOException {
	//private void serializeObject(Object o, Class clazz, String fieldName, String indent) 	throws IOException 

		Class clazz = null;
		if (o!=null)
			clazz = superClasses[superI];

		if (REENTRY_DEBUG) {
			System.err.println("-----------------------");
			if (o!=null) {
				System.err.println(o);
				System.err.println(o.getClass());
				System.err.println(clazz);
			}
		}
		//NOTE: null does not get an alias
		if (o==null)	{ //an array object can also be null...
						//in which case we lose the type; but null is only of type "null"
			printOpenTag(NULL_TOKEN, indent, false);	//null - not an object
			//no alias ID! So no call to printAliasID();
			if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);	//nice to reuse
			toOutPrintln("/>");		//good to have control of the newline here
			return;	//a null object can have no children
		}
	/*
		try {
			if (!(o instanceof Serializable))	//mark such fields "transient"
				throw new NotSerializableException(o.getClass().getName());
		} catch (Exception e) { System.err.println(e); return; }
	*/
		String runTimeClass = ParseUtilities.escapeDollar(clazz.getName());//need to: $ -> ..

//FACTOR this out into a separate method, so we can recurse nicely.
		if (clazz.isArray()) {
			serializeArrayReferences(o, fieldName, indent);
			//array is dealt with internally
		}

		else if (clazz==Class.class) {
			serializeClass((Class) o, fieldName, indent);
			//each time a string is encoded, it becomes a new object?
		}

		else if (clazz==String.class) {	//TODO: deserializer needs to use this too.
			serializeWrapper(o, fieldName, indent);
			//each time a string is encoded, it becomes a new object?
		}

		else if (clazz==Vector.class) {
			//don't check for alias, if not the leaf - but always write in full
			if ((clazz!=superClasses[0]) || putAlias(o, fieldName, false, indent)) {
			//if ((clazz!=o.getClass()) || putAlias(o, fieldName, false, indent)) 
				printOpenTag(runTimeClass, indent, true);	//putAlias, therefore object
				if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);	//nice to reuse
				if (((Vector)o).size() > 0) {
					toOutPrintln(">");		//move into ParserXML?
					serializeVectorReferences((Vector)o, indent);	//could poly
					printCloseTag(runTimeClass, indent);
				} else {
					toOutPrintln("/>");		//empty tag
				}
				if (clazz==o.getClass())	//only do for leaf (or root) classes
				//if (clazz==o.getClass())	//only do for leaf (or root) classes
					removeAlias(o);	//only acts when in "min aliases" mode
			}//putAlias
		}
		else if (clazz==Hashtable.class) {
			if ((clazz!=superClasses[0]) || putAlias(o, fieldName, false, indent)) {
			//if ((clazz!=o.getClass()) || putAlias(o, fieldName, false, indent)) 
				printOpenTag(runTimeClass, indent, true); //putAlias, therefore object
				if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);
				if (((Hashtable)o).size() > 0) {	//same as Vector
					toOutPrintln(">");		//move into ParserXML?
					serializeHashtableReferences((Hashtable)o, indent);	//poly out; nicer
					printCloseTag(runTimeClass, indent);
				} else {
					toOutPrintln("/>");		//empty tag
				}
				if (clazz==superClasses[0])	//only do for leaf classes
				//if (clazz==o.getClass())	//only do for leaf classes
					removeAlias(o);	//only acts when in "min aliases" mode
			}//putAlias
		}



		//a Class Instance
			//What if root object *is* a wrapper class?
		else {
			//8Aug2001 - we need to do writeReplace before checking the alias.
			//The replacement object is decided dynamically - we can't predict it
			//from the instance which is replaced.  From the point of view of the
			//XML stream, it is as if it the replacing object was the only one
			//that ever existed.
			//8Aug2001 - therefore, we call writeReplace now.  It's simple:
			//if (!superVersion || (superVersion && (clazz==o.getClass()))) 
			if (!superVersion || (superVersion && (clazz==superClasses[0]))) {
				o = writeReplace(o); //in case change, update info:

				if (superVersion) {
					if (o.getClass()==Object.class)	{ //special case
						superClasses = new Class[] { o.getClass() };
					} else {
						superClasses = getReversedSuperClasses(o.getClass());//leaf;no Object
					//o.getClass() is the reason for the o==null
					}
				} else {
					superClasses = new Class[] { o.getClass() };
				}
				superI = 0;
				clazz = superClasses[superI];

				//clazz = o.getClass();
				//the redundant "if" is clearer, I think
				runTimeClass = ParseUtilities.escapeDollar(clazz.getName());
			}

		 	/*  NB: we do internal first, so parser knowes when finished
				* internal serialization
				* ASSUME:
				* (0). We ignore Hashtable and Vector (use old ad hoc methods)
				* (1). defaultWriteObject() does nothing (or we get double header)
				*/
			boolean writeFull = false;
			if (superVersion) {
			  if (superHeader)	//most derived class takes role of highest superclass
					writeFull = putAlias(o, fieldName, false, indent);	//ignore result
				else
					writeFull = true; //superclasses *always* written in full
			} else {
				writeFull = putAlias(o, fieldName, false, indent);
			}
			if (writeFull) {
			//if ((clazz!=o.getClass()) || putAlias(o, fieldName, false, indent)) 
					//version=2
				if (superVersion && superI==0 && superHeader) {	//first call
					String leafClassName = ParseUtilities.escapeDollar(
						superClasses[superClasses.length-1].getName()
					);
					printOpenTag(leafClassName, indent, true);	//
					if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);	//nice to reuse
					toOutPrintln(">");		//*always* has children
								//superI always==0?
					serializeObject(false, o, superClasses, 0, SUPER_TOKEN, indent+spacer);
					//NB: same super, not header any more, no name, indented
					printCloseTag(leafClassName, indent);	//the last one
				} else {	//version=1
					printOpenTag(runTimeClass, indent, true);	//<openTag...
					if (fieldName!=null)								//...[obj-name="name"]...
						printAttr(NAME_TOKEN, fieldName);
	//test two together, to see if any dumb errors
					if (optVersion) {	//only gather here...
						serializeFields(o, clazz, indent); //starts a new objStore
							//disable this, only do when defaultWriteObject() called?
					} else {	//just reunite the bifurcated halves.
						serializeFields(o, clazz, indent); //starts a new objStore
						//  internalSer(clazz, o);	//done in serialize Fields
						serializeOptAttr(o, clazz, indent);
					}
				
						//...(field="value")*...
						//serializeFields invokes writeObject() - if any
					//firstCall is not state - just which routine calls.  Can recurse OK.
					boolean hasChildren = false;	//init
					Class superClass = null;
					if (superVersion)	{		//recursion for super classes
						//superClass = clazz.getSuperclass();
						if (superI+1<superClasses.length)
						//if (superClass!=Object.class)	//when it bottoms out
							hasChildren = true;	//only for superVersion
					}
						
					//This doesn't make sense for superVersion - it is in fact a bug
					if (!putFieldDiditAll) {  //switch on whether putFields or regular
						try {
							hasChildren = hasChildren || hasChildren(o, clazz);
							//in case superVersion
						} catch (IllegalAccessException ex) {
							throw new WriteAbortedException("", ex);//sub-IOException wrapper
						}
						if (optVersion) hasChildren = true; //opt might alone have children
						if (hasChildren) {
							toOutPrintln(">");		//move into ParserXML?
							//<fields> ----------
							if (optVersion)	//fields before opt
								serializeReferences(o, clazz, indent);//works here?  Stacking?
									//disable this, until defaultWriteObject() called?
								//(1). connect defaultWriteObject() to call serializeReferences
								//(2). need to stack the clazz and object and indent.
								// it does not work! Format has field-opt order wrong!
	
							if (INTERNAL_DEBUG) System.err.println("internal objects:");
							Vector localObjStore = objStore;//local->stack (recursion)
							Vector localDefaultCache = defaultCache;//local->stack (recursion)
	
							//<opt-data> ----------
							//if (localObjStore.size()>0) {	//make it optional (graceful degrade)	
									//this was so silly!! the objStore had not yet been filled
								boolean possibleOptData = false;	//cache
								if (optVersion) {
									//objStore = new Vector();		//but see getAllFields()...
									//defaultCache = new Vector();
									possibleOptData = doesOwnSer(clazz);
									if (possibleOptData) {
										printOpenTag(OPT_DATA_TOKEN, indent+spacer, false);//no obj
										toOutPrintln(">");	//always assume children

										if (defWO_DIRECT) {
											clazzStack.add(clazz);	//push for defWO
											thisObjectStack.add(o);
											indentStack.add(indent);
										}
										internalSer(clazz, o);	//output: obj dir + caches prim
											//push... but format requires def to be already written!
										
										if (defWO_DIRECT) {
											indentStack.remove(indentStack.size()-1);	//pop
											thisObjectStack.remove(thisObjectStack.size()-1);
											clazzStack.remove(clazzStack.size()-1);
										}

										if (primStore.size()>0) {
											printOpenTag(OPT_PRIM_DATA_TOKEN, indent+spacer+spacer, false);	//no obj
											serializeOptAttr(o, clazz, indent);
											//does the invoke (via internalSer), through false flag
											toOutPrintln("/>");	//only prim fields
										}
									}
								}
								//System.err.println("objStore==localObjStore "+(objStore==localObjStore));
								//System.err.println(localObjStore);
							if (!superVersion) {	//temp try: super to serialize direct
								for (Enumeration keys=localObjStore.elements(); keys.hasMoreElements();) {
									Object a = keys.nextElement();
									if (INTERNAL_DEBUG && a!=null) System.err.println("in objStore: "+a+"\n"+a.getClass());
									serializeObject(a, null, indent+spacer+spacer);	//debug print out
								}
							 }
								if (optVersion && possibleOptData) {
									printCloseTag(OPT_DATA_TOKEN, indent+spacer);
								}
							//}
	
								//objStore.clear(); //NB: reset when consumed
								//try: No need - now a new one each time
							if (INTERNAL_DEBUG) {
								System.err.println("should be cleared: "+localObjStore.size());
									//do internal first, so parser knows end automatically.
								System.err.println("references:");
							}
							defaultCache = localDefaultCache;	//restore! (count is OK)
							if (superVersion && superI+1<superClasses.length) {//treated like a child
								serializeObject(false, o, superClasses, superI+1, SUPER_TOKEN, indent+spacer);	//nice!!
							}
							//if (superVersion && superClass!=Object.class) //treated like a child
								//serializeObject(o, superClass, SUPER_TOKEN, indent+spacer);	//nice!!
							if (!optVersion) { //opt before fields
								// serialize out objects from PutField now.
									// each declared class, each object within each declared class
								for (int j=0; j<putFieldObjectsProtect.value().length; j++) {
									Map putFieldObjMap = putFieldObjectsProtect.value()[j];
									if (putFieldObjMap!=null) {
										for (Iterator iObj = putFieldObjMap.entrySet().iterator(); iObj.hasNext();) {
											Map.Entry pair = (Map.Entry) iObj.next();
											serializeObject(pair.getValue(), pair.getKey().toString(), indent+spacer);
												//NB: [name-value] swapped
										}
									}
								}

								serializeReferences(o, clazz, indent);
								finalizeGetAllFields();
							}
							printCloseTag(runTimeClass, indent);
						} else {
							finalizeGetAllFields();
																					//just a test, so far...
							toOutPrintln("/>");		//empty tag
						}
					} else {	//"serialPersistentFields" handled by PutField
						printCloseTag(runTimeClass, indent);//always <tag></tag>; never <tag/>
					putFieldDiditAll = false;  //out fake "pop"!
					}
				}
				//8Aug2001 - the object has now been completely written.
				//It is too late to call writeReplace.
				if (clazz==superClasses[0])	//only do for leaf classes
				//if (clazz==o.getClass())	//only do for leaf classes
					removeAlias(o);	//only acts when in "min aliases" mode
			}//putAlias
		}
	}
//NOTE: may be mixed from different objects - that's OK.
	//for defWO_DIRECT	 (need to leave in for commented out code to compile.
	Vector clazzStack = new Vector();
	Vector thisObjectStack = new Vector();
	Vector indentStack = new Vector();

	/**--------------------------------------------------------------------------
		* I wonder if these tag factorings should be in ParserXML?  It's not
		* parsing - quite the opposite - but it is at the same level as the parser.
		* In the parlace of Compilers, it is "code generation", I guess
		**/
	private void printOpenTag(String tag, String indent, boolean isObject) {
		out.print(indent + "<" +tag);
		if (isObject && includeAliasID)
			printAttr(ALIAS_ID_TOKEN, aliasSerialNumber+"");
	}
	private void printCloseTag(String tag, String indent) {
		toOutPrintln(indent + "</" +tag+ ">");
		//toOutPrintln();	//nicer formatting	hmmm... also used by Wrapper...
	}

/** ---------------------------------------------------------------------------
	* writeReplace
	* ------------
	* Reflectively invoke the present objects's writeReplace().
	* From the perspective of the XML stream, the replacement object (the one
	* that this routine returns) is the only one that ever existed.
	* "parameterless" = no-arg
	* This routine has a pattern that seems *very* similar to readResolve().
	* It may be worth factoring something out.  Defer until another 
	* reflexive invocation is required.
	**/
	Object writeReplace(Object me) throws ObjectStreamException {
		try {
			Class replacedClass = me.getClass();
			Method writeReplaceMethod = null;	//changed, for try
			try {
				writeReplaceMethod = replacedClass.getDeclaredMethod("writeReplace", new Class[]{});
			} catch (NoSuchMethodException checkSuperClasses) {	//added
				//Check for method in superclass (simulating "inheritance"):
				writeReplaceMethod = replacedClass.getMethod("writeReplace", new Class[]{});
				//To Do: iterate up the superclasses here, not just public
			}

			writeReplaceMethod.setAccessible(true);
			Object newme = writeReplaceMethod.invoke(me, new Object[]{}); //no arg
//The following check removed, because Kevin Day (Trumpet, Inc.) 12 March 2002
//showed that non-compatible replacement *is* permitted by JOS spec.
/*
			if ( !(me.getClass().isAssignableFrom(newme.getClass())) )	//ie go back
				throw new ClassCastException("The writeReplace() method of class '"+
					replacedClass.getName()	+"' returned an object of type '"+
					newme.getClass().getName() +"'.");
					//would it be more efficient to initialize all these exceptions at
					//load time?  Or better to create JIT, only as needed?
*/
			return newme;
		} catch (NoSuchMethodException e) {
			//Not an error - just means that it doesn't implement readResolve
			return me;	//if no writeReplace - the object is simply unchanged!
		} catch (IllegalAccessException e) {
			System.err.println("IllegalAccessException - this shouldn't happen, since we did setAccessible().  Perhaps security on the JVM precludes it?");
			e.printStackTrace();
			return me;	//just continue on - what would be a better thing to do?
		} catch (InvocationTargetException e) {
			System.err.println("JSX InvocationTargetException:");
			//e.printStackTrace();	//not sure how this should be handled...
			//is this the best way to handle it?  What does Serialization do?
			//Probably best to throw whatever is thrown.
			try {
				throw (ObjectStreamException) e.getTargetException(); //uncork it
				//contract of writeReplace guarantees that it only throws this class
				//However, neither the compiler nor ObjectOutputClass checks this.
				//Since this is a compile-time kind of thing, it is reasonable that
				//ObjectOutputClass does not check it, but only runtime problems...
				//Is the situation different for things like readObject() being private?
			} catch (ClassCastException f) {
				throw new Error("writeReplace() of "+me.getClass()+ " threw an '"+e.getTargetException()+"'. It should only be of type ObjectStreamException");
			}
			//unreachable - can't to get here.
		}
		//let all other Exceptions through - esp ClassCastException! - for caller
		//to handle.
	}


	/**--------------------------------------------------------------------------
		*	serializeArrayReferences
		*	------------------------
		* Can contain primitives (no wrapper) or objects
		* If called on non-array, it serializes as an Object (needed for recursion)
		**/
	private void serializeArrayReferences(Object o, String fieldName, String indent) throws IOException{
		Class clazz;
		if (o==null || !( (clazz=o.getClass()).isArray() )) {
	//		System.err.println("Just about do output the base type element of an array");
			serializeWrapperOrObject(o, indent);
			return;		//bottoms out.
		}
			//superVersion is no issue for arrays.
		if (!putAlias(o, fieldName, false, indent))
			return;	//don't inline if already done
		String runTimeClass = ParseUtilities.escapeDollar(o.getClass().getName());	//redundancy here (see above)//need to: $ -> ..
			//HUH?  The above is immediately overwritten...
		//toOutPrintln(indent + "ARRAYS NOT YET IMPLEMENTED");
		//Switch for byte[] goes here!
		if (clazz==byte[].class) {	//if a binary array, treat differently
			printOpenTag(BINARY_DATA_TOKEN, indent, true); //<binary-data -- an obj
			if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);
			printAttr(VALUE_TOKEN, ParseUtilities.encodeHex((byte[])o, formatted));
			toOutPrintln("/>");	//nice if this and the '"' were indented
			//	printCloseTag(BINARY_DATA_TOKEN, indent);	//not needed!!!
		} else {
			runTimeClass = getArrayType(clazz);
			printOpenTag(runTimeClass, indent, true); //map "[" to "ArrayOf-"; obj
			if (fieldName!=null) printAttr(NAME_TOKEN, fieldName);
			int length  = Array.getLength(o);
			printAttr(LENGTH_TOKEN, length+"");
/*
	//we comment out this section, because zero-length arrays in the emptyTag
	//format are are not being deserialized correctly.  Disabling the emptyTag
	//option is the simplest hack to fix it, as some users need it urgently.
	//Many thanks to 
	//	Raimund 'Raimi' Jacob  - Mon, 29 Jan 2001 00:47:58 +0100 (MET)
	//	Gary Lawrence Murphy  - Thu Feb  1 12:37:44 2001
	// for identifying this bug and help testing ways to solve it.
		if (length==0)		//if o.length==0
			toOutPrintln("/>");		//empty tag
		else
*/
			{		//<- this brace remains from the commented-out if-else above.
					//if primitive, put it here?
				//if *last* array (ie array of non-array), and also of a primitive.
				//well, to get here, must be an array...
				//...so check check to see if it as array of a primitive
				if ( clazz.getComponentType().isPrimitive() ) {
					for (int i=0; i<length; i++) {
						if (clazz.getComponentType() == byte.class) { //if byte[]
						 	//binary data special case. //Thanks to Mark Collette.
							//3Oct2001 - binary now a 100% drastically special case.
						} else {
							toOutPrintln();
							out.print(indent);
						}
						if (clazz.getComponentType()==char.class)	//primitive, not wrapper
							printAttr(ARRAY_PRIMITIVE_INDEX_TOKEN+i, ParseUtilities.encodeXML(Array.get(o,i)+""));
						else
							printAttr(ARRAY_PRIMITIVE_INDEX_TOKEN+i, Array.get(o,i)+"");
					}
					toOutPrintln("/>");		//*cannot* contain other tags. (into ParserXML?)
				}
				else {	//ie component is a reference type (another array or an object)
					toOutPrintln(">");		//move into ParserXML?
					for (int i=0; i<length; i++) {		//actually a recursion
						serializeArrayReferences(Array.get(o, i), null, indent+spacer);
					}
						//closeTag is inside
					printCloseTag(runTimeClass, indent);
				}
			}
		}
		removeAlias(o);
	}

	/** convert Array name string from:
		*		[*		//any number of '[' (including 0)
		*		(B|C|D|F|I|J|S|Z) |		//one of these represeting a primitive type
		*   (L[^;]*;)		//or the name of the class, in between 'L' and ';'
		**/
	static String getArrayType(Class clazz) {	//utility
		String a = "";	//will hold result - if empty string, reference problem
		String name = ParseUtilities.escapeDollar(clazz.getName());//need to: $ -> ..
		if (DEBUG) System.err.println("Array type = \""+name+"\"");
		int i;
		for (i=0; name.charAt(i)=='['; i++)
			a += ARRAY_TOKEN;
		switch (name.charAt(i++)) {
			case 'B': a += "byte"; break;
			case 'C': a += "char"; break;
			case 'D': a += "double"; break;
			case 'F': a += "float"; break;
			case 'I': a += "int"; break;
			case 'J': a += "long"; break;
			case 'S': a += "short"; break;
			case 'Z': a += "boolean"; break;
			case 'L': a += name.substring(i, name.lastIndexOf(';')); break;
						//could also be name.length()-1
		}
		// NB: in a multi-dimensional array, Class.getComponentType() returns
		// the next one further in (which may also be an array).
		return a;
	}

	/**--------------------------------------------------------------------------
		*	hasChildren
		*	-----------
		*	An inefficient but obvious prototype: better to factor out field
		* handlers (ie ref, primintive and this one), to farm out into appropriate
		* arrays for each.
		*	"Children" def: existing (ie non-null) classes (ie non-primitive).
		**/
	private boolean hasChildren(Object o, Class frameClass) throws
			IllegalAccessException, InvalidClassException {
		//System.err.println(o+"   hasChildren--------------------------");
		IllegalAccessException ex = null;	//collect as we go along
		if (objStore.size()>0)	//consider: return in fields - neat return
			return true;
		for (int j=0; j<putFieldObjectsProtect.value().length; j++) {
			Map putFieldObjMap = putFieldObjectsProtect.value()[j];
				if (putFieldObjMap!=null) {
					if (putFieldObjMap.size()>0)
						return true;
			}
		}
		//try {
			Field[] fields = getAllFields(o, frameClass, false);	//later call!
				//null indicates that we query all superclasses
			//Field[] fields = o.getClass().getFields();
			for (int i=0; i<fields.length; i++) {
				Field f = fields[i];
				if (Modifier.isStatic(f.getModifiers()))	// transient done in SerialPF
// & NOT_SERIALIZED) != 0)
					continue;
				if (!f.getType().isPrimitive()) {
					if (stringAsElement || f.getType()!=String.class) {
						//try {
							//if (f.get(o)!=null)
								return true;	// Now: we show, even if all nulls
						//} catch (IllegalAccessException e) {
						//	ex = e;	//we want to keep checking, even if an exception
						//}
					}
				}
			}
			//if no child found but exception raised, re-raise (rethrow) it.
			//(can't say for sire whether it has children or not. from Philip Gust
			//Q: What is the best feedback for the caller about this?
			if (ex != null) {
				throw ex;
			}
			return false;		//found no non-primitive (includes String),
				//non-null objects within it	
		//} catch (Exception e) {System.err.println(e); System.exit(0);}
		//get rid of this - exit() is bad in production code
	}


	/**--------------------------------------------------------------------------
		* serializeFields - attr/primitives + Strings
		* ---------------
		* SUGGEST: abstracting a "printAttr(String name, String value)" method.
		*
		*	header==true for first time
		*		<Class obj-name="..." ... >
		* header==false for second call (for opt-data)
		*   <opt-data ...>
		* Note that we have *completely* bifurcated it... (mutually exclusive)
		**/
	private void serializeFields(Object o, Class frameClass, String indent)
		throws InvalidClassException {
		//System.out.println("entering serializeFields, "+o+", "+frameClass);
		//System.err.println(o+"   serializeFields--------------------------");
		if (o==null) return;
		Field[] fields = getAllFields(o, frameClass, true);	//primitive: 1st call
			//superVersion: this needs to become based on the current frame
			//the fields should now be all from the particular frame class...
			//getAllFields invokes writeObject() - if any.
		//Field[] fields = o.getClass().getFields();
		//Field[] fields = o.getClass().getDeclaredFields();
		//getDeclaredFeilds doesn't return inherited fields.
			//NOTE!!!  getFields() returns ONLY public fields
			//getDeclaredFields() returns everything - except inherited fields
		try {
			for (int i=0; i<fields.length; i++) {
				Field f = fields[i];
				//shadowEncode not needed if (superVersion)...
				String fieldName = shadowEncode(f, o.getClass());	//added for shadowing
				//System.err.println("Field "+i+": "+f);
				if (DEBUG) System.err.println("Field "+i+": "+f);
				// if ((f.getModifiers() & NOT_SERIALIZED) != 0)
				if (Modifier.isStatic(f.getModifiers()))	// transient done in SerialPF
					continue;	//skip transient and static fields
				Class type = f.getType();
				//System.err.println("Type "+type);
						//how bad would it be to treat these two following cases the same?
					//ie. and *always* do encodeXML() automatically, in the ParserXML
				if (DEBUG) System.err.println("here1");
				if (type.isPrimitive()) {
					toOutPrintln();							//Attribute indenting
					out.print(indent);		//indent each attribute
					String outStr = f.get(o).toString();
					if (type==char.class)
						outStr = ParseUtilities.encodeXML(outStr);
					printAttr(fieldName, outStr);
				}
//Stop String being written as an attribute
				else if (!stringAsElement && f.get(o)!=null && type==String.class){	//.getName().equals("java.lang.String"))
					toOutPrintln();							//Attribute indenting
					out.print(indent);		//indent each attribute
				if (DEBUG) System.err.println("here2");
					//fieldName;
				if (DEBUG) System.err.println("here3");
					//f.get(o);
				if (DEBUG) System.err.println("here4");
					//superVersion is no issue for Strings
					if (putAlias(f.get(o), shadowEncode(f, o.getClass()), true, indent)) {
							//only inline if not done already
							//problem for attributes in String
						printAttr(fieldName, ParseUtilities.encodeXML(f.get(o)+""));
						removeAlias(f.get(o));
					}
				}
					//NB: if neither String nor primitive, it just falls right through
			}
/*
		} catch (InvocationTargetException e) {	// simple propagation
			try {
				throw (RuntimeException) e.getTargetException();	// getCause in 1.4, but need backcompat
			} catch (Exception nestedE) {	// jeez, nesting
				nestedE.printStackTrace();	// if not a runtime
			}
*/
		} catch (Exception e) {
			//System.out.println("Field: "+e);
			e.printStackTrace();
			//System.exit(1);
		}
	}


	private void serializeOptAttr(Object o, Class frameClass, String indent) {
				//hopefully the buffer stacking survives this intermediate step...
		//internalSer(c, o);
		//System.err.println("object:"+o+", "+frameClass+" about to be called");
		//if (OPT_DEBUG) System.err.println("object:"+o+", "+frameClass+" about to be called");
			//(3). process returned values.
		int i=0;
		for (Enumeration keys=primStore.elements(); keys.hasMoreElements(); i++) {
			//toOutPrintln();							//Attribute indenting
			//out.print(indent+spacer);		//indent each attribute
			printAttr(INTERNAL_PRIMITIVE_TOKEN+i, ParseUtilities.encodeXML((String)keys.nextElement()));
			//NOTE: will never be String, because it is serialized using object.
		}
		//try { throw new Exception(); }
		//catch (Exception e) { e.printStackTrace(); }

		primStore.clear();	//NB: reset when consumed
		//consider: can reuse, because never recurse. Different from objStore
		//NB: need to explicitly cast Object to String
		//Consider: if an array, could specify a String base type
		//NB: lazy for now: always encode as XML, even tho only String needs it
	}



/**--------------------------------------------------------------------------
	* isShadowed
	* ----------
	* Inherited fields with the same name as present fields are "shadowed",
	* and JSX needs some way to distinguish them.  Needed for both primitives
	* and objects (references).
	* We check to see if there is a field of the same name in a younger (sub)
	* class.  The order we scan the descendents does not matter - thus we can
	* use superclass() (no getSubclass() is avail!), and start at the bottom,
	* and work our way up until we are one class away from the class declaring
	* the field.
	* A
	* |
	* B  <-- field declared here: B.class.getDeclaredField(fieldName);>
	* |
	* C
	* |
	* D 	<-- the class of the object we have: me.getClass();
	* |
	* E
	*
	* Attribution: based on scratch code by Wes Biggs, 20 March 2001, which
	* did the opposite - if a given field shadowed another. */
	boolean isShadowed(Field f, Class c) {
		Class stopc = f.getDeclaringClass();
		for (; c!=stopc; c=c.getSuperclass()) {
  		try {
    		Field f2 = c.getDeclaredField(f.getName());
 				//System.toOutPrintln("Given field '" +qualField(f)+ "'");
 				//System.toOutPrintln("is SHADOWED BY '" +qualField(f2)+ "'");
				return true;
  		} catch (NoSuchFieldException e) {
    		// expect this in most cases
  		}
		}
 		//System.toOutPrintln("not shadowed");
		return false;
	}
/**--------------------------------------------------------------------------
	* qualField
	* ---------
	* This creates the class-qualified field name.  Arguably, this and the
	* XMLDeserialize "parseQualField" could go into the ParseUtilities class.
	**/
	String qualField(Field f) {
		return f.getDeclaringClass().getName()
				+"."+f.getName();
			//for field: $ -> ..
	}

/**--------------------------------------------------------------------------
	* shadowEncode
	* ------------
	* Returns the qualified field name, if needed; else just the plain name.
	* NOTE: ParseUtilities.escapeDollar is done when attr names are printed -
	* this is the only time it is necessary.  Note that primitive field names
	* are attributes; but reference field names are attribute *values*, and so
	* may contain "$".
	*
	* Could change this to also take into account the serialPersistentFields
	* somehow.  (1). If it was enabled; (2). the same for each of the super
	* classes.
	* I think there is a timing issue, to do with them only be set at runtime
	* (no static serialPersistentFields to declare them): one solution is to
	* just qualify them when needed to distinguish them (which might be at the
	* level of the present class).
	* 
	* The proper solution is (again) to separate the fields into those of each
	* inherited class (would this solve the runtime issue?)
	*
	**/
	String shadowEncode(Field f, Class c) {
		if (isShadowed(f, c))
			return qualField(f);
		else
			return f.getName();
			//for field: $ -> ..
			//arguable, could do all the dollar escaping here, but not as natural..
	}


	/** This should likewise be in ParserXML.
		* static!
		*/
	private void printAttr(String attrName, String value) {
		//out.print(spacer +attrName+ " = \"" +value+ "\"");
		out.print(" "+ ParseUtilities.escapeDollar(attrName)+ "=\"" +value+ "\"");
	}


	/**--------------------------------------------------------------------------
		* serializeReferences(o," ") - tags/objects
		* -------------------
		* Needs take a Class object for framing.
		**/
	//should factor out the two iterations over fields (attr and ref)
	private void serializeReferences(Object o, Class frameClass, String indent) throws IOException {
		//System.err.println(o+"  serializeReferences--------------------++++++++++++++");
		if (o==null) return;
		Field[] f = getAllFields(o, frameClass, false);	//ref: 2nd call (last call)
				//superVersion: this needs to become based on the current frame
		//Field[] f = o.getClass().getFields();
			//Field[] f = o.getClass().getDeclaredFields(); //skips inherited fields
		for (int i=0; i<f.length; i++) {
			//bug found by Romain - thanks!
			// if ((f[i].getModifiers() & NOT_SERIALIZED) != 0)//left this out before!
			if (Modifier.isStatic(f[i].getModifiers()))	// transient done in SerialPF
				continue;
			Class type = f[i].getType();
			try {
				if (!(type.isPrimitive() || (!stringAsElement && type==String.class)))		//.getName().equals("java.lang.String")))
					serializeObject(f[i].get(o), shadowEncode(f[i], o.getClass()), indent+spacer);
			} catch (Exception e) {
				System.err.println("References: " +e);
				System.err.println("i = " +i);
				System.err.println("f.length = " +f.length);
				e.printStackTrace();
			}
		}
	}


	/**--------------------------------------------------------------------------
		* getAllFields(o, frame, true)
		* ------------
		*	Walks up the class hierarchy, getting:
		*		- non-public (private etc)
		*		- inherited fields;
		* NB: getDeclaredFields() does only the former, getFields() only the latter.
		* Alt implementation:
		*		2 pass: build a vector of arrays, then arraycopy to the correct size
		* firstCall: called for both primitives and references.  But should
		* invoke internal serialization only once (on firstCall).  NB always need
		* to check for writeObject(), to know whether to skip (ie default) or not.
		**/
		Vector defaultCache;	//still is recursion
	private void finalizeGetAllFields() {
		defaultProtect.retrieve();	// pop - clear the stack
		putFieldObjectsProtect.retrieve();	// pop - clear the stack
	}
	//private Field[] getAllFields(Object o, boolean firstCall) {
	private Field[] getAllFields(Object o, Class frameClass, boolean firstCall)
		throws InvalidClassException {
				//internalSer invokes writeObject() - if any.
		Vector v = new Vector();
		int cacheIndex = 0;
		if (CACHE_DEBUG)
			System.err.println("Entering   " +firstCall);
		//if (!superVersion && firstCall) {
		if (firstCall) {
			objStore = new Vector(); 	//NB: once for all superclasses: collect all
			defaultCache = new Vector();
		}

	//Externalizable - alternative must go here...
		//NB: superVersion has no effect here
		if (o instanceof Externalizable) {
			if (firstCall) {	//only call once
				try {
					((Externalizable)o).writeExternal(thisOut);
				} catch (IOException e) {
					//should propagate, but awkward in present architecture
				}
			}
			return new Field[0];
		}

		//TEST: do just for leaf class. (later will do for all)
		//Need to go in reverse order (superclass -> subclass), because that is
		//the order that Serialization does it, and which classes which customise
		//their own serialization expect.  Thanks Mark Collette (Simulutions.com)!
		//SOLN: create an array, then step it backwards.

		//for superVersion, we ignore the whole reversed classes thing
		Class[] superClasses = null;
		if (superVersion)//with a "super" field, to nest superclasses
			superClasses = new Class[] {frameClass};
		else
			superClasses = getReversedSuperClasses(o.getClass());	//actually includes leaf...
		if (firstCall) {
			//defaultCache2 = new boolean[superClasses.length];	// create the cache
				defaultProtect.protect(new boolean[superClasses.length]);	// create the cache
				putFieldObjectsProtect.protect(new Map[superClasses.length]);	// create the cache
		}
		//System.err.println("boolean["+defaultProtect.value().length+"]");
		for (int j=0; j<superClasses.length; j++) {  //traverse root to leaf
				//c = superClasses[(superClasses.length-1)-j];	//... from root (but for Object) to leaf
			Class c = superClasses[j];	//... from root (but for Object) to leaf
			if (CACHE_DEBUG)
				System.err.print(o.getClass() +"CACHE INDEX: " +cacheIndex+"\t\t");
			if (firstCall) {	//do for each superclass
				if (!optVersion) {	//Just stop the *actual* call here
					defaultCatcher.newCatcher();
					putFieldObjectsCatcher.newCatcher();
						internalSer(c, o); //get returned values in caller (confusing?)
					putFieldObjectsProtect.value()[j] = putFieldObjectsCatcher.freeCatcher();

					boolean defaultResult = defaultCatcher.freeCatcher(); // catch and...
					defaultProtect.value()[j] = defaultResult; // ...cache
				}
				if (CACHE_DEBUG)
					System.err.println("firstcall: CACHE: " +defaultCache);
			} else {
				if (CACHE_DEBUG)
					System.err.println("not firstcall CACHE: " +defaultCache);
						//retrieve from cache for later calls
			}

			boolean doesOwn = doesOwnSer(c);
			//(1). if it doesn't do internal serialization, then serialize normally
			//(2). if it does internal serialization, only serialize if default called
				//GREAT!  Default fields omitted if defaultWriteObject() not called.
//we just completely ignore the default issue for now!
			if (optVersion ||
					(!doesOwn || (doesOwn && defaultProtect.value()[j])) ) {	//verbose is clearer
					//(!doesOwn || (doesOwn && defaultCache2[j])) ) {	//verbose is clearer
				Field[] f = c.getDeclaredFields();
				Map fmap = SerialPF.getFieldNameType(c);
				// very ineff, as we are dupicating same work
				boolean allOK = SerialPF.areSerializableFieldsValidFor_defWO(fmap, c);
				if (!allOK) throw new InvalidClassException("unmatched serializable field(s) declared");
				for (int i=0; i<f.length; i++) {
					if (fmap.containsKey(f[i].getName()))	// only if a serializable field
						v.add(f[i]);
				}
			}
		}	//old: while ( (c = c.getSuperclass()) != Object.class);	//not 'C'lass

	// The commented out, "lastCall" below:
	// we would dearly love to detect the lastcall here, and clean up the cache
	// stack here, too, but we can't tell which will be last, from here. It could
	// be from hasChildren, or from SerializeReferences.  We don't know.
	// Therefore, we delegate this clean up aspect to the code that decides
	// whether hasChildren or SerializeReferences is last.  Search on
	// "defaultProtect" to find it.
/*
		if (lastCall) {
			defaultProtect.retrieve();	// pop - clear the stack
		}
*/
	//	System.err.println("Fields for "+o.getClass()+": "+v);
		Field[] f = new Field[v.size()];	//scope of other f finished
		v.copyInto(f);		//obviate casting; and only way avail in 1.1
		//Java 1.2 security MAGIC!
		Field.setAccessible(f, true);	//new in 1.2 - need to do version checks
		return f;
	}

/**--------------------------------------------------------------------------
	* getReversedSuperClasses
	*	-----------------------
	* A utility class, common to both serialization and deserialization.
	* used by both XMLSerialize and XMLDeserialize
	**/
	static public Class[] getReversedSuperClasses(Class c) {
		if (c==Object.class)
			return new Class[0];	//Object has no superclasses
		Class storeC = c;
		boolean isObjSer = Serializable.class.isAssignableFrom(storeC);

		//int classDepth = 0;	//count for array length (faster than using Vector)
		int numSerClasses = 0;	//JOS style
		do {
			//classDepth++;
			if (!isObjSer)
				numSerClasses++; //non-Ser, JSX style (does them all)
			else if (Serializable.class.isAssignableFrom(c))
				numSerClasses++; //Ser, JOS style (don't count non-Ser)
		/* // shorter code, but less clear:
			if (isObjSer && !Serializable.class.isAssignableFrom(c))
				continue;	// skip *only* in this case
			numSerClasses++; //JOS style
		*/
		} while ( (c = c.getSuperclass()) != Object.class);		// don't count Object
		// ALT: can stop when reach first non-Ser (would inherit Ser from any
		// Ser ancestor).  Note this works for Object too (it is non-Ser).

		//if (DEBUG) System.err.println("classDepth = " + classDepth);
		//Class[] superClasses = new Class[classDepth];	//JSX style
		Class[] superClasses = new Class[numSerClasses];	//JOS style
		c = storeC; 	//restore
		int iSuper=superClasses.length-1;	//create it in reverse
		do {
			if (!isObjSer)
				superClasses[iSuper--] = c;	// non-Ser: JSX does all
			else if (Serializable.class.isAssignableFrom(c))
				superClasses[iSuper--] = c; // non-Ser: JOS skips
			//if (Serializable.class.isAssignableFrom(c))
			//	superClasses[iSuper--] = c;
		} while ( (c = c.getSuperclass()) != Object.class);
			//hmmm... nicer to start at "length", and use "--i"?  Then ends on "0"
			//in order from root to leaf:
				//[0] is the last ancestor before Object; (ie the root but for Object)
				//[last] is the leaf - the actual class of the object
		if (DEBUG) {
			for (int j=0; j<superClasses.length; j++)
				System.err.println("superClasses["+j+"] = " + superClasses[j]);
		}
		return superClasses;
	}


	/**-------------------------------------------------------------------------
		*	internalSer
		* -----------
		* returned values are stored in:
		*		- defaultFlag
		*		- objStore
		*		- primStore
		*	  - returns: whether o implements its own serialization
		* NB: With new approach to serialization, this invocation will
		* have the effect of depositing the written objects to the objStore,
		* where they will be later retrieved: a crude, one-level stack for
		*	return values.
		*
		* AH!  This is why class and object are both needed: the class forms
		* view onto the object.  Different slices are possible, for each
		* superclass...
		*
		* PutField - if writeFields was called, then don't do normal 
		* defaultWriteObject().  Actually, this *is* the standard behaviour - if
		* a class has a writeObject(), that does not explicitly call
		* defaultWriteObject, then those fields are not written out.
		* But I don't think that JSX is actually behaving in that way at the 
		* moment.
		**/
	Stack putFieldClassStack = new Stack();
		// or LinkedList: push=addLast(), pop=removeLast(), peek=getLast(), clear()
	private boolean internalSer(Class clazz, Object o) {
				//Class clazz = o.getClass();
			defaultCatcher.set(false);
		Method m;
			if ((m=getDeclaredMethod(clazz, "writeObject", OOS_ARGS,
					Modifier.PRIVATE, Modifier.STATIC)) != null) {
			if (INTERNAL_DEBUG)
				System.err.print("--JSX serialize--");
			//invokeObjectWriter(m, o);
	//NB: return values are cleared when read (consumed)
	//Consider: reset when they are not generated?
			try {
	//(2). gather (catch) returned values (in a bucket).
				if (INTERNAL_DEBUG)
					System.err.println("invoking: " +m);
				//KELLY DEBUG
/*
				System.err.println("CHECK: before writeObject: "+clazz);
				if (o.getClass()== javax.swing.JInternalFrame.JDesktopIcon.class && clazz == javax.swing.JComponent.class) {
						System.err.println("Encountered "+clazz+" in "+o.getClass()+", skipping writeObject");
				}
				else
*/
				// System.err.println(">>> about to call writeObject");
			putFieldClassStack.push(clazz);	// pushing communication
					m.invoke(o, writeObjectArglist);	//might as well do them all!
			putFieldClassStack.pop();
				// System.err.println("<<< just returned from call to writeObject");
				//objStore are in effect returned in objStore (returnedObjs?) */
				//System.err.println("invoking: " +m);
			} catch (InvocationTargetException e) {
				try {
					//System.err.println(">>> about to rethrow "+e.getTargetException());
					RuntimeException r = (RuntimeException) e.getTargetException();
					//System.err.println(">>>>>>> RuntimeException is "+r);
					throw r;
				} catch (ClassCastException nestedE) {	// not "Exception" just throw!!
					//System.err.println(">>>> can't rethrow "+e.getTargetException());
					nestedE.printStackTrace();
				}
			} catch (Exception e) {
				System.err.println(e);
				e.printStackTrace();
			}
			//NB: Only process if they exist...	hmmm, auto, if size = 0
			if (INTERNAL_DEBUG) {
			}
			if (INTERNAL_DEBUG)
				System.err.println("Object Store: "+objStore);	//check
			if (INTERNAL_DEBUG)
				System.err.println("Primitive Store: "+primStore);	//check
			/* now, we write them all out individually (not the Vector!) */
			return true; //"it implements writeObject :-)"
		}
		else {	
			if (INTERNAL_DEBUG)
				System.err.println("No writeObject() found");
			defaultCatcher.set(true);
			return false;	//"it doesn't implement writeObject :-("
		}
	}

	/**-------------------------------------------------------------------------
		*	doesOwnSer
		* ----------
 		* Need to check the same superclass three times (refs, fields, hasChild).
		* If called internalSer, would collect three lots of information!
		* We can't cache it easily, because there are typically several
		* superclasses for each object.
		* The only downside of this approach is the inefficiency.  However, the
		* way to counteract this is to refactoring the three calls into one.
		**/
	private boolean doesOwnSer(Class clazz) {
		return ((getDeclaredMethod(clazz, "writeObject", OOS_ARGS,
			Modifier.PRIVATE, Modifier.STATIC)) != null);
	}




	/**--------------------------------------------------------------------------
		* serializeVectorReferences (container type)
		* -------------------------
		* Contains objects only (primitives written as Wrappers inc String)
		* spc used instead of indent - is there any reason for this?
		*
		* Note: Does not provide extensibility by allowing subclasses to override
		* this method of Vector serialization (and Hashtable, array, etc).
		* Static methods cannot be overridden.  To provide this, a instance
		* would need to be created at some stage - perhaps lazy instantiation
		* when first called.  Then, all such method should be made object methods
		* (ie not static), so that they can be overridden.
		**/
	private void serializeVectorReferences (Vector o, String indent) throws IOException{
		final String spc = indent + spacer;
		if (o==null) throw new RuntimeException("impossible; redundant safety");
		for (Enumeration keys=o.elements(); keys.hasMoreElements(); ) {
			Object obj = keys.nextElement();
			serializeWrapperOrObject(obj, spc);
			//if (keys.hasMoreElements()) toOutPrintln();		//separating blank line
		}
	};		//what's the idea with the trailing ";"?



	/**--------------------------------------------------------------------------
		* Wrapper OR ObjectVector
		* -----------------------
		* A helper method (used for Vector and Array only so far), where the
		* object is assumed to have no "field name".
		**/
	private void serializeWrapperOrObject (Object obj, String spc) throws IOException{
		if (obj==null) {
			printOpenTag(NULL_TOKEN, spc, false);	//null is not an object
			toOutPrintln("/>");		//good to have control of the newline here
			return;
		}
		Class type = obj.getClass();
		if (type==Boolean.class ||		//is there a cleverer way to do this?
				type==Byte.class ||				//we could loop, yes, but...
				type==Short.class ||
				type==Integer.class ||
				type==Long.class ||
				type==Float.class ||
				type==Double.class)
			serializeWrapper(obj, null, spc);
		else if (type==String.class ||
				type==Character.class)
			serializeWrapper(obj, null, spc);	//will cast do Character?
		else
			serializeObject(obj, null, spc);	//null== no name
	}

	/**--------------------------------------------------------------------------
		* serializeHashtableReferences (container type)
		* ----------------------------
		* Each key and value is an object.
		* distinguished by position alone.
		* primitives type written as special "wrapper" objects (String, too).
		**/
	private void serializeHashtableReferences (
			Hashtable o, String indent) throws IOException{
		if (o==null) throw new RuntimeException("impossible; redundant safety");
		for (Enumeration keys=o.keys(); keys.hasMoreElements(); ) {
			Object key = keys.nextElement();
			Object value = o.get(key);
			Object[] objs = {key,        value};
			String[] spc =  {indent+spacer, indent+spacer};	//so can stagger, if we want
			for (int i=0; i<objs.length; i++) {
				Class type = objs[i].getClass();
				if (type==Boolean.class ||		//is there a cleverer way to do this?
						type==Byte.class ||				//we could loop, yes, but...
						type==Short.class ||
						type==Integer.class ||
						type==Long.class ||
						type==Float.class ||
						type==Double.class)
					serializeWrapper(objs[i], null, spc[i]);
				else if (type==String.class ||
						type==Character.class)
					serializeWrapper(objs[i], null, spc[i]);
				else
					serializeObject(objs[i], null, spc[i]);	//null== no name
			}
			if (keys.hasMoreElements()) toOutPrintln();		//separating blank line
		}
	}

	/** -------------------------------------------------------------------------
		* serializeWrapper()
		* ------------------
		* Note: must check aliasing, *before* encodeXML creates a new object...!
		*
		*	Is this where we should put the name, for case of root wrapper?
		* We know it must be a String in this case; but how about a more general
		*	wrapper? 
		*
		**/
	private void serializeWrapper(Object value, String name, String indent) {
		if (DEBUG_HASH) System.err.println("----enter serializeWrapper with "+name+"="+value);
		String tag = ParseUtilities.escapeDollar(value.getClass().getName());		//"java.lang.String";//need to: $ -> .. //Should *never* need this here!
			//superVersion is no issues for Wrappers.
		if (putAlias(value, name, false, indent)) { //any reason for null?
			//true returned: (first time seen - an alias was put)
			if (DEBUG_HASH) System.err.println("found NO alias, so do inline (in serializeWrapper)");
			printOpenTag(tag, indent, true);	//putAlias: therefore, an object
			if (name!=null)
				printAttr(NAME_TOKEN, name);
			if (value.getClass()==String.class) //instanceof OK, as String is final
				printAttr(VALUE_TOKEN, ParseUtilities.encodeXML((String)value));
			else if (value.getClass()==Character.class) {
				printAttr(VALUE_TOKEN, ParseUtilities.encodeXML(value+""));
			}
			else {
				printAttr(VALUE_TOKEN, value+"");
			}
			out.print("/>");		//good to have control of the newline here
				//move into ParserXML?
			//printCloseTag(tag, "");
			//NOTE: a wrapper *always* is an empty tag, because it can't have children
			toOutPrintln();	//instead of printCloseTag()
			removeAlias(value);
		}
		else {
				//false returned
			if (DEBUG_HASH) System.err.println("found object aliased already (in serializeWrapper)");
			//putAlias() takes care of this already
		}
	}
		
	private void serializeClass(Class clazz, String name, String indent) {
		String tag = "java.lang.Class";
			//superVersion is no issue for Class objects.
		if (putAlias(clazz, name, false, indent)) {
			//true returned: (first time seen - an alias was put)
			printOpenTag(tag, indent, true);	//putAlias, therefore an object
			if (name!=null)
				printAttr(NAME_TOKEN, name);
			//if (putAlias(clazz.getName(), CLASSNAME_TOKEN, true, indent))
				//comment out, because this invented field shouldn't be aliases
			printAttr(CLASSNAME_TOKEN, clazz.getName());
			out.print("/>");		//good to have control of the newline here
			toOutPrintln();	//instead of printCloseTag()
			removeAlias(clazz);
		}
	}


/** ===========================================================================
	*	===========================================================================
	* helper code for writeObject()
	*	===========================
	**/
/**--------------------------------------------------------------------------
	* getDeclaredMethod
	*	-----------------
	* A utility class, common to both serialization and deserialization
	* - need to factor out into a utility class.
	*
	* 	CHANGED: "static" - we make it static for now
	* Taken directly from ObjectOutputStream...  too general, but a good start.
	**/
	//NB: default "package" accessibility, so XMLDeserialize can access it
	static Method getDeclaredMethod(Class clazz,
			String methodName, Class[] args, 
			int requiredModifierMask,
			int disallowedModifierMask) {
		Method method = null;
		try {
			method = 
				clazz.getDeclaredMethod(methodName, args);
			if (method != null) {
				int mods = method.getModifiers();
				if ((mods & disallowedModifierMask) != 0 ||
					(mods & requiredModifierMask) != requiredModifierMask) {
						method = null;
				} else {
					method.setAccessible(true);		//needed, because prviate
				}
			}
		} catch (NoSuchMethodException e) {
	    // Since it is alright if methodName does not exist,
	    // no need to do anything special here.
		}
		return method;
	}

	/*
	try {
		if ((m=testClazz.getDeclaredMethod("writeObject", new Class[] {})) != null)
			System.err.println("JSX1 "+m+ "Class " +testClazz
				+ " has a writeObject()");
		else
			System.err.println("JSX1.  Class " +testClazz+ " has no writeObject()");
	} catch (Exception e) {
			System.err.println(e);
	}

	*/


/**--------------------------------------------------------------------------
	* invokeObjectWriter
	*	------------------
	* 	CHANGED: "static" - we make it static for now
	* 	CHANGED: send in Method m!
	* Taken directly from ObjectOutputStream... good for exceptions!
	**/
	private void invokeObjectWriter(Method m, Object obj) throws IOException {
		try {
			m.invoke(obj, writeObjectArglist);
		} catch (InvocationTargetException e) {
			Throwable t = e.getTargetException();
			if (t instanceof IOException)
				throw (IOException)t;
			else if (t instanceof RuntimeException)
				throw (RuntimeException) t;
			else if (t instanceof Error)
				throw (Error) t;
			else
				throw new Error("internal error");
		} catch (IllegalAccessException e) {
	    /* Should not happen, unless there is a problem with the VM. This
	     * is because the writeObject method is obtained via
	     * ObjectStreamClass.getDeclaredMethod, which calls setAccessible
	     * on it, in privileged mode.
	     */
			throw new InternalError("Unable to access writeObject method");
		}
	}

/**--------------------------------------------------------------------------
  * Formatted flag
  * true if the output is to be formatted
  */
  private boolean formatted = true; //defaults to formatted
  //private boolean formatted = false;	//easy test

/**--------------------------------------------------------------------------
  * Spacer String
  * Set according to the formatted value
  */
  private String spacer = SPACER;	//defaults to formatted
  //private String spacer = "";				//NB: need to execute code to set this;
		//or change all references to it.
  
/**--------------------------------------------------------------------------
  *  setFormatted(boolean)
  *  Turns XML output formatting on or off...
  **/
  public void setFormatted(boolean formatted) {
	  this.formatted = formatted;
	  spacer = (formatted?SPACER:"");
  }
  
/**--------------------------------------------------------------------------
  *  toOutPrint(String s)
  **/
  private void toOutPrint(String s) {
    out.print(s);
  }
/**--------------------------------------------------------------------------
  *  toOutPrintln()
  **/
  private void toOutPrintln() {
    if (formatted)
      out.println();
  }
/**--------------------------------------------------------------------------
  *  toOutPrintln(String s)
  **/
  private void toOutPrintln(String s) {
    if (formatted)
      out.println(s);
    else
      out.print(s);
  }
}
